from os import path as pth

import numpy as np

from astropy import units as u
from astropy.table import Table
from astropy.io import fits
from astropy.utils.decorators import deprecated_renamed_argument

from synphot import SourceSpectrum, ConstFlux1D, Empirical1D
from synphot.units import PHOTLAM

from ..optics import image_plane_utils as ipu
from .source_utils import make_img_wcs_header
from .source import Source
from .. import rc

__all__ = ["empty_sky", "star", "star_field"]


def empty_sky(flux=0):
    """
    Returns an empty source so that instrumental fluxes can be simulated

    Returns
    -------
    sky : Source

    """
    sky = Source(lam=[0.3, 3.0]*u.um, spectra=[flux, flux]*PHOTLAM,
                 x=[0], y=[0], ref=[0], weight=[1])
    return sky


@deprecated_renamed_argument('mag', 'flux', '0.1.5')
def star(x=0, y=0, flux=0):
    """
    Source object for a single star in either vega, AB magnitudes, or Jansky

    The star is associated with the reference spectrum for each photometric
    system, therefore a reference wavelength or filter does not need to be given

    Parameters
    ----------
    x, y : float
        [arcsec] position from centre of field of view
    flux : float
        [vega mag, AB mag, Jy] Stellar brightness

    Returns
    -------
    src : Source
        A source object with a single entry table field and a reference spectrum

    """

    mag_unit = u.mag
    spec_template = vega_spectrum
    if isinstance(flux, u.Quantity):
        if flux.unit.physical_type == "spectral flux density":  # ABmag and Jy
            mag_unit = u.ABmag
            spec_template = ab_spectrum
            flux = flux.to(u.ABmag)
        flux = flux.value

    spec = spec_template()

    w = 10**(-0.4 * flux)
    ref = 0

    tbl = Table(data=[[x], [y], [w], [ref], [flux]],
                names=["x", "y", "weight", "ref", "mag"],
                units=[u.arcsec, u.arcsec, None, None, mag_unit])
    tbl.meta["photometric_system"] = "vega" if mag_unit == u.mag else "ab"
    src = Source(spectra=spec, table=tbl)
    src.meta.update({"function_call": f"star(x={x}, y={y}, flux={flux})",
                     "module": "scopesim.source.source_templates",
                     "object": "star"})

    return src


def star_field(n, mmin, mmax, width, height=None, use_grid=False):
    """
    Creates a super basic field of stars with random positions and brightnesses

    Parameters
    ----------
    n : int
        number of stars

    mmin, mmax : float, astropy.Quantity
        [mag, ABmag, Jy] min and max magnitudes/fluxes of the population stars.
        If floats, then assumed Quantity is vega magnitudes

    width : float
        [arcsec] width of region to put stars in

    height : float, optional
        [arcsec] if None, then height=width

    use_grid : bool, optional
        Place stars randomly or on a grid

    Returns
    -------
    stars : scopesim.Source object
        A Source object with a field of stars that can be fed into the method:
        OpticalTrain.observe()

    See Also
    --------
    OpticalTrain.observe
    OpticalTrain.readout

    """
    if height is None:
        height = width

    mag_unit = u.mag
    spec_template = vega_spectrum
    if isinstance(mmin, u.Quantity):
        if mmin.unit.physical_type == "spectral flux density":  # ABmag and Jy
            mag_unit = u.ABmag
            spec_template = ab_spectrum
            mmin, mmax = mmin.to(u.ABmag), mmax.to(u.ABmag)
        mmin, mmax = mmin.value, mmax.value

    spec = spec_template()

    if rc.__config__["!SIM.random.seed"] is not None:
        np.random.seed(rc.__config__["!SIM.random.seed"])

    if use_grid:
        nw = np.ceil(n**0.5)
        nh = np.ceil(n / nw)
        x, y = np.mgrid[0:1:1/nw, 0:1:1/nh] - 0.5
        positions = x.flatten(), y.flatten()
    else:
        positions = np.random.random(size=(2, n)) - 0.5

    x = width * positions[0]
    y = height * positions[1]
    mags = np.linspace(mmin, mmax, n)
    w = 10**(-0.4 * mags)
    ref = np.zeros(n, dtype=int)

    tbl = Table(data=[x, y, w, ref, mags],
                names=["x", "y", "weight", "ref", "mag"],
                units=[u.arcsec, u.arcsec, None, None, mag_unit])
    tbl.meta["photometric_system"] = "vega" if mag_unit == u.mag else "ab"
    stars = Source(spectra=spec, table=tbl)

    return stars


def uniform_illumination(xs, ys, pixel_scale, flux=None, spectrum=None):
    """
    Return a Source for a uniformly illuminated area

    Parameters
    ----------
    xs, ys : list of float
        [arcsec] min and max extent of each dimension relative to FOV centre
        E.g. `xs=[-1, 1], ys=[5, 5.5]`

    pixel_scale : float
        [arcsec]

    flux : astropy.Quantity
        [mag, ABMag, Jy] Flux per arcsecond of the Source

    Returns
    -------
    src : scopesim.Source

    Examples
    --------
    A 200x200 uniform illumination Source at 1 Jy/arcsec2
    ::

        src = uniform_illumination(xs=[-1,1], ys=[-1, 1],
                                   pixel_scale=0.01, flux=1*u.Jy)


    A source that extends just past the MICADO 15" slit dimensions with a flux
    of 10 mag/arcsec2
    ::

        src = uniform_illumination(xs=[-8, 8], ys=[-0.03, 0.03],
                                   pixel_scale=0.004, flux=10*u.mag)

    Using a self made frequency-comb spectrum with 1 Jy lines ever 0.1µm
    ::

        import numpy as np
        from astropy import units as u
        from synphot import SourceSpectrum, Empirical1D

        wave = np.arange(0.7, 2.5, 0.001) * u.um
        flux = np.zeros(len(wave))
        flux[::100] = 1 * u.Jy
        spec = SourceSpectrum(Empirical1D, points=wave, lookup_table=flux)

        src = uniform_illumination(xs=[-8, 8], ys=[-0.03, 0.03],
                                   pixel_scale=0.004, spectrum=spec)

    """
    if flux is not None:
        mag_unit = u.mag
        spec_template = vega_spectrum
        if isinstance(flux, u.Quantity):
            if flux.unit.physical_type == "spectral flux density":  # ABmag and Jy
                mag_unit = u.ABmag
                spec_template = ab_spectrum
                flux = flux.to(u.ABmag)
            flux = flux.value

        spec = spec_template()
        scale_factor = pixel_scale ** 2 * 10 ** (-0.4 * flux)
    elif spectrum is not None:
        spec = spectrum
        mag_unit = "PHOTLAM"
        scale_factor = pixel_scale ** 2
    else:
        raise ValueError(f"Either flux or spectrum must be passed: {flux}, {spectrum}")

    hdr = ipu.header_from_list_of_xy(x=np.array(xs) / 3600.,
                                     y=np.array(ys) / 3600.,
                                     pixel_scale=pixel_scale / 3600.)

    data = scale_factor * np.ones([max(1, hdr["NAXIS2"]), max(1, hdr["NAXIS1"])])
    hdu = fits.ImageHDU(header=hdr, data=data)

    hdu.header["SPEC_REF"] = 0
    hdu.header["SPECMAG"] = 0
    hdu.header["SPECUNIT"] = str(mag_unit)

    src = Source(image_hdu=hdu, spectra=[spec])

    return src


def uniform_source(sp=None, extent=60):
    """
    Simplified form of scopesim_templates.misc.uniform_source, mostly intended for testing

    This function creates an image with extend^2 pixels with pixel size of 1 arcsec^2 so provided amplitudes
    are in flux or magnitudes per arcsec^2

    It accepts any synphot.SourceSpectrum compatible object

    sp : synphot.SourceSpectrum
         defaults to vega_spectrum() with magnitude 0 mag/arcsec2

    extent : int, default 60
        extension of the field in arcsec, will always produce a square field. Default value produces a field of 60x60 arcsec

    """
    if sp is None:
        sp = vega_spectrum()

    data = np.ones(shape=(extent, extent))

    header = make_img_wcs_header(pixel_scale=1, image_size=data.shape)
    hdu = fits.ImageHDU(header=header, data=data)

    src = Source(spectra=sp, image_hdu=hdu)

    return src


def vega_spectrum(mag=0):
    if isinstance(mag, u.Quantity):
        mag = mag.value
    vega = SourceSpectrum.from_file(pth.join(rc.__pkg_dir__, "vega.fits"))
    vega = vega * 10 ** (-0.4 * mag)
    return vega


def st_spectrum(mag=0):
    # ..todo: the waves vector is a bit random, in particular its length, but sets the resolution of
    #         the final spectrum in scopesim. Can this be make more general?
    waves = np.geomspace(100, 300000, 50000)
    sp = ConstFlux1D(amplitude=mag*u.STmag)

    return SourceSpectrum(Empirical1D, points=waves, lookup_table=sp(waves))


def ab_spectrum(mag=0):
    # ..todo: the waves vector is a bit random, in particular its length, but sets the resolution of
    #         the final spectrum in scopesim. Can this be make more general?
    waves = np.geomspace(100, 300000, 50000)
    sp = ConstFlux1D(amplitude=mag * u.ABmag)

    return SourceSpectrum(Empirical1D, points=waves, lookup_table=sp(waves))



import numpy as np
from scipy import ndimage as spi
from scipy.interpolate import RectBivariateSpline, griddata
from scipy.ndimage import zoom
from astropy import units as u
from astropy.convolution import Gaussian2DKernel
from astropy.io import fits
import matplotlib.pyplot as plt
from matplotlib.colors import LogNorm

from .. import rc, utils
from ..optics import image_plane_utils as imp_utils


def round_kernel_edges(kernel):
    y, x = np.array(kernel.shape).astype(int) // 2
    threshold = np.min([kernel[y, 0], kernel[y, -1],
                        kernel[0, x], kernel[-1, x]])
    kernel[kernel < threshold] = 0.

    return kernel


def nmrms_from_strehl_and_wavelength(strehl, wavelength, strehl_hdu,
                                     plot=False):
    """
    Returns the wavefront error needed to make a PSF with a desired strehl ratio

    Parameters
    ----------
    strehl : float
        [0.001, 1] Desired strehl ratio. Values 1<sr<100 will be scale to <1
    wavelength : float
        [um]
    strehl_hdu : np.ndarray
        2D map of strehl ratio as a function of wavelength [um] and residual
        wavefront error [nm RMS]
    plot : bool

    Returns
    -------
    nm : float
        [nm] residual wavefront error for generating an on-axis AnisoCADO PSF
        with the desired strehl ratio at a given wavelength

    """

    if 1. < strehl < 100.:
        strehl *= 0.01

    nm0 = strehl_hdu.header["CRVAL1"]
    dnm = strehl_hdu.header["CDELT1"]
    nm1 = strehl_hdu.header["NAXIS1"] * dnm + nm0
    nms = np.arange(nm0, nm1, dnm)

    w0 = strehl_hdu.header["CRVAL2"]
    dw = strehl_hdu.header["CDELT2"]
    w1 = strehl_hdu.header["NAXIS2"] * dw + w0
    ws = np.arange(w0, w1, dw)

    strehl_map = strehl_hdu.data

    nms_spline = RectBivariateSpline(ws, nms, strehl_map, kx=1, ky=1)
    strehls = nms_spline(wavelength, nms)[0]

    if strehl > np.max(strehls):
        raise ValueError("Strehl ratio ({}) is impossible at this wavelength "
                         "({}). Maximum Strehl possible is {}."
                         "".format(strehl, wavelength, np.max(strehls)))

    if strehls[0] < strehls[-1]:
        nm = np.interp(strehl, strehls, nms)
    else:
        nm = np.interp(strehl, strehls[::-1], nms[::-1])

    if plot:
        plt.plot(nms, strehls)
        plt.plot(nm, strehl, "ro")
        plt.show()

    return nm


def make_strehl_map_from_table(tbl, pixel_scale=1*u.arcsec):


    # pixel_scale = utils.quantify(pixel_scale, u.um).to(u.deg)
    # coords = np.array([tbl["x"], tbl["y"]]).T
    #
    # xmin, xmax = np.min(tbl["x"]), np.max(tbl["x"])
    # ymin, ymax = np.min(tbl["y"]), np.max(tbl["y"])
    # mesh = np.array(np.meshgrid(np.arange(xmin, xmax, pixel_scale),
    #                             np.arange(np.min(tbl["y"]), np.max(tbl["y"]))))
    # map = griddata(coords, tbl["layer"], mesh, method="nearest")
    #

    map = griddata(np.array([tbl.data["x"], tbl.data["y"]]).T,
                   tbl.data["layer"],
                   np.array(np.meshgrid(np.arange(-25, 26),
                                        np.arange(-25, 26))).T,
                   method="nearest")

    hdr = imp_utils.header_from_list_of_xy(np.array([-25, 25]) / 3600.,
                                           np.array([-25, 25]) / 3600.,
                                           pixel_scale=1/3600)

    map_hdu = fits.ImageHDU(header=hdr, data=map)

    return map_hdu


def rescale_kernel(image, scale_factor, spline_order=None):
    if spline_order is None:
        spline_order = utils.from_currsys("!SIM.computing.spline_order")
    sum_image = np.sum(image)
    image = zoom(image, scale_factor, order=spline_order)
    image = np.nan_to_num(image, copy=False)        # numpy version >=1.13

    # Re-centre kernel
    im_shape = image.shape
    dy, dx = np.divmod(np.argmax(image), im_shape[1]) - np.array(im_shape) // 2
    if dy > 0:
        image = image[2*dy:, :]
    elif dy < 0:
        image = image[:2*abs(dy), :]
    if dx > 0:
        image = image[:, 2*dx:]
    elif dx < 0:
        image = image[:, 2*abs(dx):]

    sum_new_image = np.sum(image)
    image *= sum_image / sum_new_image

    return image


def cutout_kernel(image, fov_header):
    h, w = image.shape
    xcen, ycen = 0.5 * w, 0.5 * h
    dx = 0.5 * fov_header["NAXIS1"]
    dy = 0.5 * fov_header["NAXIS2"]
    x0, x1 = max(0, int(xcen-dx)), min(w, int(xcen+dx))
    y0, y1 = max(0, int(ycen-dy)), min(w, int(ycen+dy))
    image_cutout = image[y0:y1, x0:x1]

    return image_cutout


def get_strehl_cutout(fov_header, strehl_imagehdu):

    image = np.zeros((fov_header["NAXIS2"], fov_header["NAXIS1"]))
    canvas_hdu = fits.ImageHDU(header=fov_header, data=image)
    canvas_hdu = imp_utils.add_imagehdu_to_imagehdu(strehl_imagehdu,
                                                    canvas_hdu, spline_order=0,
                                                    conserve_flux=False)
    canvas_hdu.data = canvas_hdu.data.astype(int)

    return canvas_hdu


def nearest_index(x, x_array):
    # return int(round(np.interp(x, x_array, np.arange(len(x_array)))))
    return np.argmin(abs(x_array - x))


def get_psf_wave_exts(hdu_list, wave_key="WAVE0"):
    """
    Returns a dict of {extension : wavelength}

    Parameters
    ----------
    hdu_list

    Returns
    -------
    wave_set, wave_ext

    """

    if not isinstance(hdu_list, fits.HDUList):
        raise ValueError("psf_effect must be a PSF object: {}"
                         "".format(type(hdu_list)))

    tmp = np.array([[ii, hdu.header[wave_key]]
                    for ii, hdu in enumerate(hdu_list)
                    if wave_key in hdu.header and hdu.data is not None])
    wave_ext = tmp[:, 0].astype(int)
    wave_set = tmp[:, 1]

    # ..todo:: implement a way of getting the units from WAVEUNIT
    # until then assume everything is in um
    wave_set = utils.quantify(wave_set, u.um)

    return wave_set, wave_ext


def get_total_wfe_from_table(tbl):
    wfes = utils.quantity_from_table("wfe_rms", tbl, "um")
    n_surfs = tbl["n_surfaces"]
    total_wfe = np.sum(n_surfs * wfes**2)**0.5

    return total_wfe


def wfe2gauss(wfe, wave, width=None):
    strehl = wfe2strehl(wfe, wave)
    sigma = strehl2sigma(strehl)
    if width is None:
        width = int(np.ceil(8 * sigma))
        width += (width + 1) % 2
    gauss = sigma2gauss(sigma, x_size=width, y_size=width)

    return gauss


def wfe2strehl(wfe, wave):
    wave = utils.quantify(wave, u.um)
    wfe = utils.quantify(wfe, u.um)
    x = 2 * 3.1415926526 * wfe / wave
    strehl = np.exp(-x**2)
    return strehl


def strehl2sigma(strehl):
    amplitudes = [0.00465, 0.00480, 0.00506, 0.00553, 0.00637, 0.00793, 0.01092,
                  0.01669, 0.02736, 0.04584, 0.07656, 0.12639, 0.20474, 0.32156,
                  0.48097, 0.66895, 0.84376, 0.95514, 0.99437, 0.99982, 0.99999]
    sigmas = [19.9526, 15.3108, 11.7489, 9.01571, 6.91830, 5.30884, 4.07380,
              3.12607, 2.39883, 1.84077, 1.41253, 1.08392, 0.83176, 0.63826,
              0.48977, 0.37583, 0.28840, 0.22130, 0.16982, 0.13031, 0.1]
    sigma = np.interp(strehl, amplitudes, sigmas)
    return sigma


def sigma2gauss(sigma, x_size=15, y_size=15):
    kernel = Gaussian2DKernel(sigma, x_size=x_size, y_size=y_size,
                              mode="oversample").array
    kernel /= np.sum(kernel)
    return kernel


def rotational_blur(image, angle):
    """
    Rotates and coadds an image over a given angle to imitate a blur

    Parameters
    ----------
    image : array
        Image to blur
    angle : float
        [deg] Angle over which the image should be rotationally blurred

    Returns
    -------
    image_rot : array
        Blurred image

    """
    image_rot = np.copy(image)

    n_angles = 1
    rad_to_deg = 57.29578
    edge_pixel_unit_angle = np.arctan2(1, (image.shape[0] // 2)) * rad_to_deg
    while abs(angle) > edge_pixel_unit_angle and n_angles < 25:
        angle /= 2.
        image_rot += spi.rotate(image_rot, angle, reshape=False, order=1)
        # each time kernel is rotated and added, the frame total doubles
        n_angles *= 2

    return image_rot / n_angles

def get_bkg_level(obj, bg_w):
    """
    Determine the background level of image or cube slices

    Returns a scalar if obj is a 2d image or a vector if obj is a 3D cube (one
    value for each plane).
    The method for background determination is decided by self.meta["bkg_width"]:
    If 0, the background is returned as zero (implying no background subtraction).
    If -1, the background is estimated as the median of the entire image (or
    cube plane).
    If positive, the background is estimated as the median of a frame of width
    `bkg_width` around the edges.
    """

    if obj.ndim == 2:
        if bg_w == 0:
            bkg_level = 0
        else:
            mask = np.zeros_like(obj, dtype=np.bool8)
            if bg_w > 0:
                mask[bg_w:-bg_w,bg_w:-bg_w] = True
            bkg_level = np.ma.median(np.ma.masked_array(obj, mask=mask))

    elif obj.ndim == 3:
        if bg_w == 0:
            bkg_level = np.array([0] * obj.shape[0])
        else:
            mask = np.zeros_like(obj, dtype=np.bool8)
            if bg_w > 0:
                mask[:, bg_w:-bg_w, bg_w:-bg_w] = True
            bkg_level = np.ma.median(np.ma.masked_array(obj, mask=mask),
                                     axis=(2, 1)).data

    else:
        raise ValueError("Unsupported dimension:", obj.ndim)
    return bkg_level



from .image_plane import ImagePlane
from . import image_plane_utils

from .optical_element import OpticalElement
from .optical_train import OpticalTrain
from .optics_manager import OpticsManager

from . import spectrograph

from .surface import SpectralSurface
from . import surface_utils
from .radiometry import RadiometryTable
from . import radiometry_utils

from .fov import FieldOfView
from .fov_manager import FOVManager


import os
import sys
from copy import deepcopy
from shutil import copyfileobj

from datetime import datetime

import numpy as np
from scipy.interpolate import interp1d
from astropy.io import fits
from astropy.wcs import WCS
from astropy import units as u

from synphot.units import PHOTLAM

from .optics_manager import OpticsManager
from .fov_manager import FOVManager
from .image_plane import ImagePlane
from ..commands.user_commands import UserCommands
from ..detector import DetectorArray
from ..effects import ExtraFitsKeywords
from ..source.source import Source
from ..utils import from_currsys
from ..version import version
from .. import effects
from .. import rc
from . import fov_utils as fu


class OpticalTrain:
    """
    The main class for controlling a simulation

    Parameters
    ----------
    cmds : UserCommands, str
        If the name of an instrument is passed, OpticalTrain tries to find the
        instrument package, and internally creates the UserCommands object

    Examples
    --------
    Create an optical train::

        >>> import scopesim as im
        >>> cmd = sim.UserCommands("MICADO")
        >>> opt = sim.OpticalTrain(cmd)

    Observe a Source object::

        >>> src = sim.source.source_templates.empty_sky()
        >>> opt.observe(src)
        >>> hdus = opt.readout()

    List the effects modelled in an OpticalTrain::

        >>> print(opt.effects)

    Effects can be accessed by using the name of the effect::

        >>> print(opt["dark_current"])

    To include or exclude an effect during a simulation run, use the
    ``.include`` attribute of the effect::

         >>> opt["dark_current"].include = False

    Data used by an Effect object is contained in the ``.data`` attribute, while
    other information is contained in the ``.meta`` attribute::

        >>> opt["dark_current"].data
        >>> opt["dark_current"].meta

    Meta data values can be set by either using the ``.meta`` attribute
    directly::

        >>> opt["dark_current"].meta["value"] = 0.5

    or by passing a dictionary (with one or multiple entries) to the
    OpticalTrain object::

        >>> opt["dark_current"] = {"value": 0.75, "dit": 30}

    """
    def __init__(self, cmds=None):

        self._description = self.__repr__()
        self.cmds = cmds
        self.optics_manager = None
        self.fov_manager = None
        self.image_planes = []
        self.detector_arrays = []
        self.yaml_dicts = None
        self._last_source = None

        if cmds is not None:
            self.load(cmds)

    def load(self, user_commands):
        """
        (Re)Loads an OpticalTrain with a new set of UserCommands

        Parameters
        ----------
        user_commands : UserCommands

        """

        if isinstance(user_commands, str):
            user_commands = UserCommands(use_instrument=user_commands)

        if not isinstance(user_commands, UserCommands):
            raise ValueError("user_commands must be a UserCommands object: "
                             "{}".format(type(user_commands)))

        self.cmds = user_commands
        rc.__currsys__ = user_commands
        self.yaml_dicts = rc.__currsys__.yaml_dicts
        self.optics_manager = OpticsManager(self.yaml_dicts)
        self.update()

    def update(self, **kwargs):
        """
        Update the user-defined parameters and remake the main internal classes

        Parameters
        ----------
        kwargs : expanded dict
            Any keyword-value pairs from a config file

        """
        self.optics_manager.update(**kwargs)
        opt_man = self.optics_manager

        self.fov_manager = FOVManager(opt_man.fov_setup_effects, **kwargs)
        self.image_planes = [ImagePlane(hdr, **kwargs)
                             for hdr in opt_man.image_plane_headers]
        self.detector_arrays = [DetectorArray(det_list, **kwargs)
                                for det_list in opt_man.detector_setup_effects]


    def observe(self, orig_source, update=True, **kwargs):
        """
        Main controlling method for observing ``Source`` objects

        Parameters
        ----------
        orig_source : Source
        update : bool
            Reload optical system
        kwargs : expanded dict
            Any keyword-value pairs from a config file

        Notes
        -----
        How the list of Effects is split between the 5 main tasks:

        - Make a FOV list - z_order = 0..99
        - Make a image plane - z_order = 100..199
        - Apply Source altering effects - z_order = 200..299
        - Apply FOV specific (3D) effects - z_order = 300..399
        - Apply FOV-independent (2D) effects - z_order = 400..499
        - [Apply detector plane (0D, 2D) effects - z_order = 500..599]

        .. todo:: List is out of date - update

        """
        if update:
            self.update(**kwargs)

        self.set_focus(**kwargs)    # put focus back on current instrument package

        # Make a copy of the Source and prepare for observation (convert to
        # internally used units, sample to internal wavelength grid)
        source = orig_source.make_copy()
        source = self.prepare_source(source)

        # [1D - transmission curves]
        for effect in self.optics_manager.source_effects:
            source = effect.apply_to(source)

        # [3D - Atmospheric shifts, PSF, NCPAs, Grating shift/distortion]
        fovs = self.fov_manager.fovs
        for fov in fovs:
            # print("FOV", fov_i+1, "of", n_fovs, flush=True)
            # .. todo: possible bug with bg flux not using plate_scale
            #          see fov_utils.combine_imagehdu_fields
            fov.extract_from(source)

            hdu_type = "cube" if self.fov_manager.is_spectroscope else "image"
            fov.view(hdu_type)
            for effect in self.optics_manager.fov_effects:
                fov = effect.apply_to(fov)

            fov.flatten()
            self.image_planes[fov.image_plane_id].add(fov.hdu, wcs_suffix="D")
            # ..todo: finish off the multiple image plane stuff

        # [2D - Vibration, flat fielding, chopping+nodding]
        for effect in self.optics_manager.image_plane_effects:
            for ii in range(len(self.image_planes)):
                self.image_planes[ii] = effect.apply_to(self.image_planes[ii])

        self._last_fovs = fovs
        self._last_source = source


    def prepare_source(self, source):
        """
        Prepare source for observation

        The method is currently applied to cube fields only.
        The source data are converted to internally used units (PHOTLAM).
        The source data are interpolated to the waveset used by the FieldOfView
        This is necessary when the source data are sampled on a coarser grid
        than used internally, or if the source data are sampled on irregular
        wavelengths.
        For cube fields, the method assumes that the wavelengths at which the
        cube is sampled is provided explicitely as attribute `wave` if the cube
        ImageHDU.
        """
        # Convert to PHOTLAM per arcsec2
        # ..todo: this is not sufficiently general

        for cube in source.cube_fields:
            header, data, wave = cube.header, cube.data, cube.wave

            # Need to check whether BUNIT is per arcsec2 or per pixel
            inunit = u.Unit(header['BUNIT'])
            data = data.astype(np.float32) * inunit
            factor = 1
            for base, power in zip(inunit.bases, inunit.powers):
                if (base**power).is_equivalent(u.arcsec**(-2)):
                    conversion =(base**power).to(u.arcsec**(-2)) / base**power
                    data *= conversion
                    factor = u.arcsec**(-2)

            data = data.to(PHOTLAM,
                           equivalencies=u.spectral_density(wave[:, None, None]))

            if factor == 1:    # Normalise to 1 arcsec2 if not a spatial density
                # ..todo: lower needed because "DEG" is not understood, this is ugly
                pixarea = (header['CDELT1'] * u.Unit(header['CUNIT1'].lower()) *
                           header['CDELT2'] * u.Unit(header['CUNIT2'].lower())).to(u.arcsec**2)
                data = data / pixarea.value    # cube is per arcsec2

            data = (data * factor).value

            cube.header['BUNIT'] = 'PHOTLAM/arcsec2'    # ..todo: make this more explicit?

            # The imageplane_utils like to have the spatial WCS in units of "deg". Ensure
            # that the cube is passed on accordingly
            cube.header['CDELT1'] = header['CDELT1'] * u.Unit(header['CUNIT1'].lower()).to(u.deg)
            cube.header['CDELT2'] = header['CDELT2'] * u.Unit(header['CUNIT2'].lower()).to(u.deg)
            cube.header['CUNIT1'] = 'deg'
            cube.header['CUNIT2'] = 'deg'

            # Put on fov wavegrid
            wave_min = min([fov.meta["wave_min"] for fov in self.fov_manager.fovs])
            wave_max = max([fov.meta["wave_max"] for fov in self.fov_manager.fovs])
            wave_unit = u.Unit(from_currsys("!SIM.spectral.wave_unit"))
            dwave = from_currsys("!SIM.spectral.spectral_bin_width")  # Not a quantity
            fov_waveset = np.arange(wave_min.value, wave_max.value, dwave) * wave_unit
            fov_waveset = fov_waveset.to(u.um)

            # Interpolate into new data cube.
            # This is done layer by layer for memory reasons.
            new_data = np.zeros((fov_waveset.shape[0], data.shape[1], data.shape[2]),
                                dtype=np.float32)
            for j in range(data.shape[1]):
                cube_interp = interp1d(wave.to(u.um).value, data[:, j, :],
                                       axis=0, kind="linear",
                                       bounds_error=False, fill_value=0)
                new_data[:, j, :] = cube_interp(fov_waveset.value)

            cube.data = new_data
            cube.header['CTYPE3'] = 'WAVE'
            cube.header['CRPIX3'] = 1
            cube.header['CRVAL3'] = wave_min.value
            cube.header['CDELT3'] = dwave
            cube.header['CUNIT3'] = wave_unit.name

        return source

    def readout(self, filename=None, **kwargs):
        """
        Produces detector readouts for the observed image

        Parameters
        ----------
        filename : str, optional
            Where to save the FITS file
        kwargs

        Returns
        -------
        hdu : fits.HDUList

        Notes
        -----
        - Apply detector plane (0D, 2D) effects - z_order = 500..599

        """

        hduls = []
        for i, detector_array in enumerate(self.detector_arrays):
            array_effects = self.optics_manager.detector_array_effects
            dtcr_effects = self.optics_manager.detector_effects
            hdul = detector_array.readout(self.image_planes, array_effects,
                                         dtcr_effects, **kwargs)

            fits_effects = self.optics_manager.get_all(ExtraFitsKeywords)
            if len(fits_effects) > 0:
                for effect in fits_effects:
                    hdul = effect.apply_to(hdul, optical_train=self)
            else:
                try:
                    hdul = self.write_header(hdul)
                except Exception as error:
                    print("\nWarning: header update failed, data will be saved with incomplete header.")
                    print("Reason: ", sys.exc_info()[0], error)
                    print("")

            if filename is not None and isinstance(filename, str):
                fname = filename
                if len(self.detector_arrays) > 1:
                    fname = str(i) + "_" + filename
                hdul.writeto(fname, overwrite=True)

            hduls += [hdul]

        return hduls

    def write_header(self, hdulist):
        """Writes meaningful header to simulation product"""

        # Primary hdu
        pheader = hdulist[0].header
        pheader['DATE'] = datetime.now().isoformat(timespec='seconds')
        pheader['ORIGIN'] = 'Scopesim ' + version
        pheader['INSTRUME'] = from_currsys("!OBS.instrument")
        pheader['INSTMODE'] = ", ".join(from_currsys("!OBS.modes"))
        pheader['TELESCOP'] = from_currsys("!TEL.telescope")
        pheader['LOCATION'] = from_currsys("!ATMO.location")

        # Source information taken from first only.
        # ..todo: What if source is a composite?
        srcfield = self._last_source.fields[0]
        if type(srcfield).__name__ == "Table":
            pheader['SOURCE'] = "Table"
        elif type(srcfield).__name__ == "ImageHDU":
            if 'BG_SURF' in srcfield.header:
                pheader['SOURCE'] = srcfield.header['BG_SURF']
            else:
                try:
                    pheader['SOURCE'] = srcfield.header['FILENAME']
                except KeyError:
                    pheader['SOURCE'] = "ImageHDU"

        # Image hdul
        # ..todo: currently only one, update for detector arrays
        # ..todo: normalise filenames - some need from_currsys, some need os.path.basename
        #         this should go into a function so as to reduce clutter here.
        iheader = hdulist[1].header
        iheader['EXPTIME'] = from_currsys("!OBS.exptime"), "[s]"
        iheader['DIT'] = from_currsys("!OBS.dit"), "[s]"
        iheader['NDIT'] = from_currsys("!OBS.ndit")
        iheader['BUNIT'] = 'e', 'per EXPTIME'
        iheader['PIXSCALE'] = from_currsys("!INST.pixel_scale"), "[arcsec]"

        # A simple WCS
        iheader['CTYPE1'] = 'LINEAR'
        iheader['CTYPE2'] = 'LINEAR'
        iheader['CRPIX1'] = (iheader['NAXIS1'] + 1) / 2
        iheader['CRPIX2'] = (iheader['NAXIS2'] + 1) / 2
        iheader['CRVAL1'] = 0.
        iheader['CRVAL2'] = 0.
        iheader['CDELT1'] = iheader['PIXSCALE']
        iheader['CDELT2'] = iheader['PIXSCALE']
        iheader['CUNIT1'] = 'arcsec'
        iheader['CUNIT2'] = 'arcsec'

        for eff in self.optics_manager.detector_setup_effects:
            efftype = type(eff).__name__

            if efftype == "DetectorList" and eff.include:
                iheader['DETECTOR'] = eff.meta['detector']

        for eff in self.optics_manager.detector_array_effects:
            efftype = type(eff).__name__

            if (efftype == "DetectorModePropertiesSetter" and
                eff.include):
                # ..todo: can we write this into currsys?
                iheader['DET_MODE'] = (eff.meta['detector_readout_mode'],
                                       "detector readout mode")
                iheader['MINDIT'] = from_currsys("!DET.mindit"), "[s]"
                iheader['FULLWELL'] = from_currsys("!DET.full_well"), "[s]"
                iheader['RON'] = from_currsys("!DET.readout_noise"), "[e]"
                iheader['DARK'] = from_currsys("!DET.dark_current"), "[e/s]"

        ifilter = 1   # Counts filter wheels
        isurface = 1  # Counts surface lists
        for eff in self.optics_manager.source_effects:
            efftype = type(eff).__name__

            if efftype == "ADCWheel" and eff.include:
                iheader['ADC'] = eff.current_adc.meta['name']

            if efftype == "FilterWheel" and eff.include:
                iheader[f'FILTER{ifilter}'] = (eff.current_filter.meta['name'],
                                               eff.meta['name'])
                ifilter += 1

            if efftype == "SlitWheel" and eff.include:
                iheader['SLIT'] = (eff.current_slit.meta['name'],
                                   eff.meta['name'])

            if efftype == "PupilTransmission" and eff.include:
                iheader['PUPTRANS'] = (from_currsys("!OBS.pupil_transmission"),
                                       "cold stop, pupil transmission")

            if efftype == "SkycalcTERCurve" and eff.include:
                iheader['ATMOSPHE'] = "Skycalc", "atmosphere model"
                iheader['LOCATION'] = eff.meta['location']
                iheader['AIRMASS'] = eff.meta['airmass']
                iheader['TEMPERAT'] = eff.meta['temperature'], '[degC]'
                iheader['HUMIDITY'] = eff.meta['humidity']
                iheader['PRESSURE'] = eff.meta['pressure'], '[hPa]'
                iheader['PWV'] = eff.meta['pwv'], "precipitable water vapour"

            if efftype == "AtmosphericTERCurve" and eff.include:
                iheader['ATMOSPHE'] = eff.meta['filename'], "atmosphere model"
                # ..todo: expand if necessary

            if efftype == "SurfaceList" and eff.include:
                iheader[f'SURFACE{isurface}'] = (eff.meta['filename'],
                                                 eff.meta['name'])
                isurface += 1

            if efftype == "QuantumEfficiencyCurve" and eff.include:
                iheader['QE'] = os.path.basename(eff.meta['filename']), eff.meta['name']

        for eff in self.optics_manager.fov_effects:
            efftype = type(eff).__name__

            # ..todo: needs to be handled with isinstance(eff, PSF)
            if efftype == "FieldConstantPSF" and eff.include:
                iheader["PSF"] = eff.meta['filename'], "point spread function"

            if efftype == "SpectralTraceList" and eff.include:
                iheader["SPECTRAC"] = (from_currsys(eff.meta['filename']),
                                       "spectral trace definition")
                if "CTYPE1" in eff.meta:
                    for key in ['WCSAXES', 'CTYPE1', 'CTYPE2', 'CRPIX1', 'CRPIX2', 'CRVAL1',
                                'CRVAL2', 'CDELT1', 'CDELT2', 'CUNIT1', 'CUNIT2']:
                        iheader[key] = eff.meta[key]

        for eff in self.optics_manager.detector_effects:
            efftype = type(eff).__name__

            if efftype == "LinearityCurve" and eff.include:
                iheader['DETLIN'] = from_currsys(eff.meta['filename'])

        return hdulist

    def set_focus(self, **kwargs):
        self.cmds.update(**kwargs)
        dy = self.cmds.default_yamls
        if len(dy) > 0 and "packages" in dy:
            self.cmds.update(packages=self.default_yamls[0]["packages"])
        rc.__currsys__ = self.cmds


    def shutdown(self):
        """
        Shut down the instrument.

        This method closes all open file handles and should be called when the optical train
        is no longer needed.
        """
        for effect_name in self.effects['name']:
            try:
                self[effect_name]._file.close()
            except AttributeError:
                pass

        self._description = "The instrument has been shut down."


    @property
    def effects(self):
        return self.optics_manager.list_effects()

    def __str__(self):
        return self._description

    def __getitem__(self, item):
        return self.optics_manager[item]

    def __setitem__(self, key, value):
        self.optics_manager[key] = value

    def report(self):
        pass
        # user commands report
        #   package dependencies
        #   modes names
        #   default modes
        #   yaml hierarchy
        #
        # optics_manager
        #   derived properties
        #   system transmission curve
        #   list of effects
        #
        # etc
        #   limiting magnitudes
        #



import os
from copy import deepcopy
from ... import UserCommands, OpticalTrain
from ... import rc


def load_example_optical_train(**kwargs):
    """
    Returns an basic example ``OpticalTrain`` object with IMG and SPEC modes

    Parameters
    ----------
    Any of the additional parameters taken by UserCommands.
    E.g. ``properties``, ``set_modes``

    Examples
    --------
    ::

        import scopesim as sim
        from scopesim.source import source_templates as st

        src = st.star()
        opt = sim.load_example_optical_train(set_modes=["spectroscopy"])
        opt["filter_wheel"].change_filter("J")

        opt.observe(src)
        hdul = opt.readout()[0]


    Returns
    -------
    opt: OpticalTrain

    """
    # old_local_path = deepcopy(rc.__config__["!SIM.file.local_packages_path"])
    # old_search_path = deepcopy(rc.__search_path__)

    inst_pkgs = os.path.dirname(__file__)
    if inst_pkgs not in rc.__search_path__:
        rc.__search_path__ = rc.__search_path__ + [inst_pkgs]
        rc.__config__["!SIM.file.local_packages_path"] = inst_pkgs

    cmd = UserCommands(use_instrument="basic_instrument", **kwargs)
    opt = OpticalTrain(cmd)

    # rc.__config__["!SIM.file.local_packages_path"] = old_local_path
    # rc.__search_path__ = old_search_path

    return opt



import numpy as np
from astropy import units as u
import matplotlib.pyplot as plt

from .ter_curves import TERCurve
from ..optics import radiometry_utils as rad_utils
from ..optics.surface import PoorMansSurface, SpectralSurface
from ..utils import quantify, from_currsys


class SurfaceList(TERCurve):
    def __init__(self, **kwargs):
        super(SurfaceList, self).__init__(**kwargs)
        params = {"z_order": [20, 120, 520],
                  "minimum_throughput": "!SIM.spectral.minimum_throughput",
                  "etendue": "!TEL.etendue",
                  "report_plot_include": True,
                  "report_table_include": True}
        self.meta.update(params)
        self.meta.update(kwargs)

        self.surfaces = rad_utils.make_surface_dict_from_table(self.table)
        self._surface = None
        self._throughput = None
        self._emission = None

    def fov_grid(self, which="waveset", **kwargs):
        if which == "waveset":
            self.meta.update(kwargs)
            self.meta = from_currsys(self.meta)
            wave_min = quantify(self.meta["wave_min"], u.um)
            wave_max = quantify(self.meta["wave_max"], u.um)
            # ..todo:: add 1001 to default.yaml somewhere
            wave = np.linspace(wave_min, wave_max, 1001)
            throughput = self.throughput(wave)
            threshold = self.meta["minimum_throughput"]
            valid_waves = np.where(throughput >= threshold)[0]
            if len(valid_waves) > 0:
                wave_edges = [min(wave[valid_waves]), max(wave[valid_waves])]
            else:
                raise ValueError("No transmission found above the threshold {} "
                                 "in this wavelength range {}. Did you open "
                                 "the shutter?"
                                 "".format(self.meta["minimum_throughput"],
                                           [self.meta["wave_min"],
                                            self.meta["wave_max"]]))
        else:
            wave_edges = []

        return wave_edges

    @property
    def throughput(self):
        if self._throughput is None:
            self._throughput = self.get_throughput()
        return self._throughput

    @property
    def emission(self):
        if "etendue" not in self.meta:
            raise ValueError("self.meta['etendue'] must be set")
        etendue = quantify(from_currsys(self.meta["etendue"]), "m2 arcsec2")
        if self._emission is None:
            self._emission = self.get_emission(etendue)

        return self._emission

    @property
    def surface(self):
        if self._surface is None:
            self._surface = PoorMansSurface(self.emission, self.throughput,
                                            self.meta)

        return self._surface

    @surface.setter
    def surface(self, item):
        self._surface = item

    def get_throughput(self, start=0, end=None, rows=None):
        """ Copied directly from radiometry_table """

        if self.table is None:
            return None
        end = len(self.table) if end is None else end
        end = end + len(self.table) if end < 0 else end
        rows = np.arange(start, end) if rows is None else rows

        thru = rad_utils.combine_throughputs(self.table, self.surfaces, rows)

        return thru

    def get_emission(self, etendue, start=0, end=None, rows=None,
                     use_area=False):
        # ..todo:: work out what this use_area flag means!

        if self.table is None:
            return None
        end = len(self.table) if end is None else end
        end = end + len(self.table) if end < 0 else end
        rows = np.arange(start, end) if rows is None else rows

        emission = rad_utils.combine_emissions(self.table, self.surfaces, rows,
                                               etendue, use_area)

        return emission

    @property
    def area(self):
        areas = [0] + [self.surfaces[key].area.to(u.m**2).value
                       for key in self.surfaces
                       if self.surfaces[key].area is not None]
        return np.max(areas) * u.m**2

    @property
    def is_empty(self):
        return len(self.table) == 0

    # .. todo:: remove: Relic of the old SurfaceList
    def add_surface(self, surface, name=None, position=-1, add_to_table=True):
        if name is None:
            name = surface.meta.get("name", "<unknown surface>")
        if isinstance(surface, TERCurve):
            ter_meta = surface.meta
            surface = surface.surface
            surface.meta.update(ter_meta)

        self.surfaces.update({name: surface})
        self.table = rad_utils.add_surface_to_table(self.table, surface,
                                                    name, position)

    # .. todo:: remove: Relic of the old SurfaceList
    def add_surface_list(self, surface_list, prepend=False):
        if isinstance(surface_list, SurfaceList):
            self.surfaces.update(surface_list.surfaces)
            self.table = rad_utils.combine_tables(surface_list.table,
                                                  self.table, prepend)

    def plot(self, which="x", wavelength=None, ax=None, **kwargs):
        """

        Parameters
        ----------
        which : str
            "x" plots throughput. "t","e","r" plot trans/emission/refl
        wavelength
        kwargs

        Returns
        -------

        """
        import matplotlib.pyplot as plt
        plt.gcf().clf()

        for ii, which_part in enumerate(which):
            ax = plt.subplot(len(which), 1, ii+1)

            # Plot the individual surfaces
            for key in self.surfaces:
                ter = TERCurve(**self.surfaces[key].meta)
                ter.surface = self.surfaces[key]
                ter.plot(which=which_part, wavelength=None, ax=None,
                         new_figure=False,
                         plot_kwargs={"ls": "-", "label": key}, **kwargs)

            # Plot the system surface
            ter = TERCurve(**self.meta)
            ter.surface = self.surface
            ter.plot(which=which_part, wavelength=None, ax=ax, new_figure=False,
                     plot_kwargs={"ls": "-.", "label": "System Throughput"},
                     **kwargs)

        plt.legend()

        return plt.gcf()



import logging
from inspect import isclass

from astropy.table import Table

from .. import effects as efs
from ..effects.effects_utils import make_effect, get_all_effects
from ..utils import write_report
from ..reports.rst_utils import table_to_rst
from .. import rc


class OpticalElement:
    """
    Contains all information to describe a section of an optical system

    There are 5 major section: ``location``, ``telescope``, ``relay optics``
    ``instrument``, ``detector``.

    An OpticalElement describes how a certain section of the optical train
    changes the incoming photon distribution by specifying a list of
    ``Effects`` along with a set of local properties e.g. temperature, etc
    which are common to more than one Effect


    Parameters
    ----------
    yaml_dict : dict
        Description of optical section properties, effects, and meta-data

    kwargs : dict
        Optical Element specific information which has no connection to the
        effects that are passed. Any global values, e.g. airmass
        (i.e. bang strings) are passed on to the individual effect, where the
        relevant values are then pulled from rc.__currsys__


    Attributes
    ----------
    meta : dict
        Contains meta data from the yaml, and is updated with the OBS_DICT
        key-value pairs

    properties : dict
        Contains any properties that is "global" to the optical element, e.g.
        instrument temperature, atmospheric pressure. Any OBS_DICT keywords are
        cleaned with from the ``meta`` dict during initialisation

    effects : list of dicts
        Contains the a list of dict descriptions of the effects that the optical
        element generates. Any OBS_DICT keywords are cleaned with from the
        ``meta`` dict during initialisation


    Examples
    --------

    """
    def __init__(self, yaml_dict=None, **kwargs):
        self.meta = {"name": "<empty>"}
        self.meta.update(kwargs)
        self.properties = {}
        self.effects = []

        if isinstance(yaml_dict, dict):
            self.meta.update({key: yaml_dict[key] for key in yaml_dict
                              if key not in ["properties", "effects"]})
            if "properties" in yaml_dict:
                self.properties = yaml_dict["properties"]
            if "name" in yaml_dict:
                self.properties["element_name"] = yaml_dict["name"]
            if "effects" in yaml_dict and len(yaml_dict["effects"]) > 0:
                for eff_dic in yaml_dict["effects"]:
                    if "name" in eff_dic and hasattr(rc.__currsys__,
                                                     "ignore_effects"):
                        if eff_dic["name"] in rc.__currsys__.ignore_effects:
                            eff_dic["include"] = False

                    self.effects += [make_effect(eff_dic, **self.properties)]

    def add_effect(self, effect):
        if isinstance(effect, efs.Effect):
            self.effects += [effect]
        else:
            logging.warning("{} is not an Effect object and was not added"
                          "".format(effect))

    def get_all(self, effect_class):
        return get_all_effects(self.effects, effect_class)

    def get_z_order_effects(self, z_level):
        if isinstance(z_level, int):
            zmin = z_level
            zmax = zmin + 99
        elif isinstance(z_level, (tuple, list)):
            zmin, zmax = z_level[:2]
        else:
            zmin, zmax = 0, 999

        effects = []
        for eff in self.effects:
            if eff.include and "z_order" in eff.meta:
                z = eff.meta["z_order"]
                if isinstance(z, (list, tuple)):
                    if any([zmin <= zi <= zmax for zi in z]):
                        effects += [eff]
                else:
                    if zmin <= z <= zmax:
                        effects += [eff]

        return effects

    @property
    def surfaces_list(self):
        _ter_list = [effect for effect in self.effects
                     if isinstance(effect, (efs.SurfaceList, efs.FilterWheel,
                                            efs.TERCurve))]
        return _ter_list

    @property
    def masks_list(self):
        _mask_list = [effect for effect in self.effects if
                      isinstance(effect, (efs.ApertureList, efs.ApertureMask))]
        return _mask_list

    def list_effects(self):
        elements = [self.meta["name"]] * len(self.effects)
        names = [eff.display_name for eff in self.effects]
        classes = [eff.__class__.__name__ for eff in self.effects]
        included = [eff.meta["include"] for eff in self.effects]
        z_orders = [eff.meta["z_order"] for eff in self.effects]

        colnames = ["element", "name", "class", "included", "z_orders"]
        data = [elements, names, classes, included, z_orders]
        tbl = Table(names=colnames, data=data, copy=False)

        return tbl

    def __add__(self, other):
        self.add_effect(other)

    def __getitem__(self, item):
        """
        Returns Effects of Effect meta properties

        Parameters
        ----------
        item : str, int, Effect-Class
            Either the name, list index, or Class of an effect.
            It is possible to get meta entries from a named effect by using:
            ``#<effect_name>.<meta-key>``

        Returns
        -------
        obj : str, float, Effect, list
            - str, float : if #-string is used to get Effect.meta entries
            - Effect : if a unique Effect name is given
            - list : if an Effect-Class is given

        Examples
        --------
        ::
            from scopesim.effect import TERCurve
            opt_el = opt_elem.OpticalElement(detector_yaml_dict)

            effect_list = opt_el[TERCurve]
            effect = opt_el[0]
            effect = opt_el["detector_qe_curve"]
            meta_value = opt_el["#detector_qe_curve.filename"]

        """
        obj = None
        if isclass(item):
            obj = self.get_all(item)
        elif isinstance(item, int):
            obj = self.effects[item]
        elif isinstance(item, str):
            if item[0] == "#" and "." in item:
                eff, meta = item.replace("#", "").split(".")
                obj = self[eff][f"#{meta}"]
            else:
                obj = [eff for eff in self.effects
                       if eff.meta["name"] == item]

        if isinstance(obj, list) and len(obj) == 1:
            obj = obj[0]
        # if obj is None or len(obj) == 0:
        #     logging.warning(f'No result for key: "{item}". '
        #                     f'Did you mean "#{item}"?')

        return obj

    def __repr__(self):
        msg = '\nOpticalElement : "{}" contains {} Effects: \n' \
              ''.format(self.meta["name"], len(self.effects))
        eff_str = "\n".join(["[{}] {}".format(i, eff.__repr__())
                             for i, eff in enumerate(self.effects)])
        return msg + eff_str

    def __str__(self):
        name = self.meta.get("name", self.meta.get("filename", "<empty>"))
        return '{}: "{}"'.format(type(self).__name__, name)

    @property
    def properties_str(self):
        prop_str = ""
        max_key_len = max([len(key) for key in self.properties.keys()])
        for key in self.properties:
            if key not in ["comments", "changes", "description", "history",
                           "report"]:
                prop_str += "    {} : {}\n".format(key.rjust(max_key_len),
                                                   self.properties[key])

        return prop_str

    def report(self, filename=None, output="rst", rst_title_chars="^#*+",
               **kwargs):

        rst_str = """
{}
{}

**Element**: {}

**Alias**: {}
        
**Description**: {}

Global properties
{}
::

{}
""".format(str(self),
           rst_title_chars[0] * len(str(self)),
           self.meta.get("object", "<unknown optical element>"),
           self.meta.get("alias", "<unknown alias>"),
           self.meta.get("description", "<no description>"),
           rst_title_chars[1] * 17,
           self.properties_str)

        if len(self.list_effects()) > 0:
            rst_str += """        
Effects
{}

Summary of Effects included in this optical element:

.. table::
    :name: {}
   
{}
 
""".format(rst_title_chars[1] * 7,
           "tbl:" + self.meta.get("name", "<unknown OpticalElement>"),
           table_to_rst(self.list_effects(), indent=4))

        reports = [eff.report(rst_title_chars=rst_title_chars[-2:], **kwargs)
                   for eff in self.effects]
        rst_str += "\n\n" + "\n\n".join(reports)

        write_report(rst_str, filename, output)

        return rst_str



from copy import deepcopy
import numpy as np

from ..base_classes import ImagePlaneBase, DetectorBase
from ..optics import image_plane_utils as imp_utils
from .. import utils

from astropy.io import fits


class Detector(DetectorBase):
    def __init__(self, header, **kwargs):
        image = np.zeros((header["NAXIS2"], header["NAXIS1"]))
        self._hdu = fits.ImageHDU(header=header, data=image)
        self.meta = {}
        self.meta.update(header)
        self.meta.update(kwargs)

    def extract_from(self, image_plane, spline_order=1, reset=True):
        if reset:
            self.reset()
        if not isinstance(image_plane, ImagePlaneBase):
            raise ValueError("image_plane must be an ImagePlane object: {}"
                             "".format(type(image_plane)))

        self._hdu = imp_utils.add_imagehdu_to_imagehdu(image_plane.hdu,
                                                       self.hdu, spline_order,
                                                       wcs_suffix="D")

    def reset(self):
        self._hdu.data = np.zeros(self._hdu.data.shape)

    @property
    def hdu(self):
        new_meta = utils.stringify_dict(self.meta)
        self._hdu.header.update(new_meta)
        return self._hdu

    @property
    def header(self):
        return self._hdu.header

    @property
    def data(self):
        return self._hdu.data

    @property
    def image(self):
        return self.data



import logging

import numpy as np
from astropy import units as u, wcs
from astropy.io import fits
from astropy.table import Table
from scipy.ndimage import interpolation as ndi

from .. import utils


###############################################################################
# Make Canvas


def get_canvas_header(hdu_or_table_list, pixel_scale=1 * u.arcsec):
    """
    Generate a fits.Header with a WCS that covers everything in the FOV

    Parameters
    ----------
    hdu_or_table_list : list
        A list of Tables and/or ImageHDU py_objects

    pixel_scale : astropy.Quantity
        [arcsec] The pixel scale of the projection. Default in 1 arcsec

    Returns
    -------
    header : fits.Header
        A Header containing a WCS and NAXISn values to build an ImageHDU

    """

    size_warning = "Header dimension are {} large: {}. Any image made from " \
                   "this header will use more that >{} in memory"

    headers = [ht.header for ht in hdu_or_table_list
               if isinstance(ht, fits.ImageHDU)]
    if sum([isinstance(ht, Table) for ht in hdu_or_table_list]) > 0:
        tbls = [ht for ht in hdu_or_table_list if isinstance(ht, Table)]
        tbl_hdr = _make_bounding_header_for_tables(tbls,
                                                   pixel_scale=pixel_scale)
        headers += [tbl_hdr]

    if len(headers) > 0:
        hdr = _make_bounding_header_from_imagehdus(headers,
                                                   pixel_scale=pixel_scale)
        num_pix = hdr["NAXIS1"] * hdr["NAXIS2"]
        if num_pix > 2 ** 25:  # 2 * 4096**2
            logging.warning(size_warning.format("", num_pix, "256 MB"))
        elif num_pix > 2 ** 28:
            raise MemoryError(size_warning.format("too", num_pix, "8 GB"))
    else:
        logging.warning("No tables or ImageHDUs were passed")
        hdr = None

    return hdr


def _make_bounding_header_from_imagehdus(imagehdus, pixel_scale=1*u.arcsec):
    """
    Returns a Header with WCS and NAXISn keywords bounding all input ImageHDUs

    Parameters
    ----------
    imagehdus : list of fits.ImageHDU
    pixel_scale : u.Quantity
        [arcsec]

    Returns
    -------
    hdr : fits.Header

    """

    x = []
    y = []
    if pixel_scale.unit.physical_type == "angle":
        s = ""
    elif pixel_scale.unit.physical_type == "length":
        s = "D"

    for imagehdu in imagehdus:
        if isinstance(imagehdu, fits.ImageHDU):
            x_foot, y_foot = calc_footprint(imagehdu.header, s)
        else:
            x_foot, y_foot = calc_footprint(imagehdu, s)
        x += list(x_foot)
        y += list(y_foot)
    unit = u.mm if s == "D" else u.deg
    pixel_scale = pixel_scale.to(unit).value
    hdr = header_from_list_of_xy(x, y, pixel_scale, s)
    hdr["NAXIS1"] += 2
    hdr["NAXIS2"] += 2

    return hdr


def _make_bounding_header_for_tables(tables, pixel_scale=1*u.arcsec):
    """
    Returns a Header with WCS and NAXISn keywords bounding all input Tables

    Parameters
    ----------
    tables : list of astropy.Tables
        [arcsec] Must contain columns: "x", "y"

    pixel_scale : u.Quantity
        [arcsec]

    Returns
    -------
    hdr : fits.Header

    """

    x = []
    y = []

    s = "D" if pixel_scale.unit.physical_type == "length" else ""
    unit_new = u.mm if s == "D" else u.deg
    unit_orig = u.mm if s == "D" else u.arcsec
    x_name = "x_mm" if s == "D" else "x"
    y_name = "y_mm" if s == "D" else "y"

    pixel_scale = pixel_scale.to(unit_new).value
    for table in tables:
        x_col = utils.quantity_from_table(x_name, table, unit_orig).to(unit_new)
        y_col = utils.quantity_from_table(y_name, table, unit_orig).to(unit_new)
        x_col = list(x_col.value)
        y_col = list(y_col.value)
        x += [np.min(x_col), np.max(x_col) + 2 * pixel_scale]
        y += [np.min(y_col), np.max(y_col) + 2 * pixel_scale]

    hdr = header_from_list_of_xy(x, y, pixel_scale, s)

    return hdr


def header_from_list_of_xy(x, y, pixel_scale, wcs_suffix=""):
    """
    Makes a header large enough to contain all x,y on-sky coordinates

    Parameters
    ----------
    x, y : list of floats
        [deg, mm] List of sky coordinates to be bounded by the NAXISn keys
    pixel_scale : float
        [deg, mm]

    Returns
    -------
    hdr : fits.Header

    """

    s = wcs_suffix
    if wcs_suffix != "D":
        x = np.array(x)
        x[x > 270] -= 360
        x[x <= -90] += 360
        x = list(x)

    hdr = fits.Header()

    # .. todo: find out how this plays with chunks
    # crval1 = divmod(min(x), pixel_scale)[0] * pixel_scale
    # crval2 = divmod(min(y), pixel_scale)[0] * pixel_scale

    # naxis1 = int((max(x) - crval1) // pixel_scale) # + 2
    # naxis2 = int((max(y) - crval2) // pixel_scale) # + 2

    crval1 = min(x)
    crval2 = min(y)

    # ..todo:: test whether abs(pixel_scale) breaks anything
    pixel_scale = abs(pixel_scale)
    dx = (max(x) - min(x)) / pixel_scale
    dy = (max(y) - min(y)) / pixel_scale
    naxis1 = int(np.round(dx))
    naxis2 = int(np.round(dy))

    hdr["NAXIS"] = 2
    hdr["NAXIS1"] = naxis1
    hdr["NAXIS2"] = naxis2
    hdr["CTYPE1"+s] = "LINEAR" if s == "D" else "RA---TAN"
    hdr["CTYPE2"+s] = "LINEAR" if s == "D" else "DEC--TAN"
    hdr["CUNIT1"+s] = "mm" if s == "D" else "deg"
    hdr["CUNIT2"+s] = "mm" if s == "D" else "deg"
    hdr["CDELT1"+s] = pixel_scale
    hdr["CDELT2"+s] = pixel_scale
    hdr["CRVAL1"+s] = crval1
    hdr["CRVAL2"+s] = crval2
    hdr["CRPIX1"+s] = 0.
    hdr["CRPIX2"+s] = 0.

    xpcen, ypcen = naxis1 // 2, naxis2 // 2
    xscen, yscen = pix2val(hdr, xpcen, ypcen, s)
    hdr["CRVAL1"+s] = float(xscen)
    hdr["CRVAL2"+s] = float(yscen)
    hdr["CRPIX1"+s] = xpcen
    hdr["CRPIX2"+s] = ypcen

    return hdr


###############################################################################
# Table overlays


def add_table_to_imagehdu(table, canvas_hdu, sub_pixel=True, wcs_suffix=""):
    """
    Add files from an astropy.Table to the image of an fits.ImageHDU

    Parameters
    ----------
    table : astropy.Table
        Must contain the columns "x_mm", "y_mm", "flux" with the units in the
        column attribute .unit, or in the table.meta dictionary as
        "<colname>_unit". Default units are ``mm`` and ``ph / s / pix``

    canvas_hdu : fits.ImageHDU
        The ImageHDU onto which the table files should be projected.
        This must include a valid WCS

    sub_pixel : bool, optional
        Default is True. If True, sub-pixel shifts of files will be taken into
        account when projecting onto the canvas pixel grid. This takes about 5x
        longer than ignoring the sub-pixel shifts

    wcs_suffix : str, optional

    Returns
    -------
    canvas_hdu : fits.ImageHDU

    """

    s = wcs_suffix
    if not utils.has_needed_keywords(canvas_hdu.header, s):
        raise ValueError("canvas_hdu must include an appropriate WCS: {}"
                         "".format(s))

    f = utils.quantity_from_table("flux", table, default_unit=u.Unit("ph s-1"))
    if s == "D":
        x = utils.quantity_from_table("x_mm", table, default_unit=u.mm).to(u.mm)
        y = utils.quantity_from_table("y_mm", table, default_unit=u.mm).to(u.mm)
    else:
        arcsec = u.arcsec
        x = utils.quantity_from_table("x", table, default_unit=arcsec).to(u.deg)
        y = utils.quantity_from_table("y", table, default_unit=arcsec).to(u.deg)

    xpix, ypix = val2pix(canvas_hdu.header, x.value, y.value, s)

    # Weird FITS / astropy behaviour. Axis1 == y, Axis2 == x.
    naxis1 = canvas_hdu.header["NAXIS1"]
    naxis2 = canvas_hdu.header["NAXIS2"]
    # Also occasionally 0 is returned as ~ -1e-11
    eps = -1e-7
    mask = (xpix >= eps) * (xpix < naxis1) * (ypix >= eps) * (ypix < naxis2)

    if sub_pixel is True:
        canvas_hdu = _add_subpixel_sources_to_canvas(canvas_hdu, xpix, ypix, f,
                                                     mask)
    else:
        canvas_hdu = _add_intpixel_sources_to_canvas(canvas_hdu, xpix, ypix, f,
                                                     mask)

    return canvas_hdu


def _add_intpixel_sources_to_canvas(canvas_hdu, xpix, ypix, flux, mask):
    canvas_hdu.header["comment"] = "Adding {} int-pixel files" \
                                   "".format(len(flux))
    xpix = xpix.astype(int)
    ypix = ypix.astype(int)
    for ii in range(len(xpix)):
        if mask[ii]:
            canvas_hdu.data[ypix[ii], xpix[ii]] += flux[ii].value

    return canvas_hdu


def _add_subpixel_sources_to_canvas(canvas_hdu, xpix, ypix, flux, mask):
    canvas_hdu.header["comment"] = "Adding {} sub-pixel files" \
                                   "".format(len(flux))
    canvas_shape = canvas_hdu.data.shape
    for ii in range(len(xpix)):
        if mask[ii]:
            xx, yy, fracs = sub_pixel_fractions(xpix[ii], ypix[ii])
            for x, y, frac in zip(xx, yy, fracs):
                if y < canvas_shape[0] and x < canvas_shape[1]:
                    canvas_hdu.data[y, x] += frac * flux[ii].value

    return canvas_hdu


def sub_pixel_fractions(x, y):
    """
    Makes a list of pixel coordinates and weights to reflect sub-pixel shifts

    A point source which isn't centred on a pixel can be modelled by a centred
    PSF convolved with a shifted delta function. A fraction of the delta
    function in moved into each of the adjoining pixels. For example, a star
    at ``(x,y)=(0.2, 0.2)`` would be represented by a following pixel weights::

       ---------------
       | 0.16 | 0.04 |
       ---------------
       | 0.64 | 0.16 |
       ---------------

    where (0,0) is the centre of the bottom-left pixel

    Given (x,y) pixel coordinates, this function returns the fractions of flux
    that should go into the surrounding pixels, as well as the coordinates of
    those neighbouring pixels.

    Parameters
    ----------
    x, y : float

    Returns
    -------
    x_pix, y_pix, fracs : list of (int, int, float)
       The x and y pixel coordinates and their corresponding flux fraction

    """

    x0, dx = divmod(x, 1)
    y0, dy = divmod(y, 1)

    xi0 = int(x0)
    xi1 = xi0 + bool(dx)
    yi0 = int(y0)
    yi1 = yi0 + bool(dy)

    f00 = (1. - dx) * (1. - dy)
    f01 = (1. - dx) * dy
    f10 = dx * (1 - dy)
    f11 = dx * dy

    x_pix = [xi0, xi1, xi0, xi1]
    y_pix = [yi0, yi0, yi1, yi1]
    fracs = [f00, f10, f01, f11]

    return x_pix, y_pix, fracs


###############################################################################
# Image overlays
#
#
# def overlay_image_old(small_im, big_im, coords, mask=None, sub_pixel=False):
#     """
#     Overlay small_im on top of big_im at the position specified by coords
#
#     ``small_im`` will be centred at ``coords``
#
#     Adapted from:
#     ``https://stackoverflow.com/questions/14063070/
#         overlay-a-smaller-image-on-a-larger-image-python-opencv``
#
#     """
#
#     # TODO - Add in a catch for sub-pixel shifts
#     if sub_pixel:
#         raise NotImplementedError
#
#     x, y = np.array(coords, dtype=int) - np.array(small_im.shape) // 2
#
#     # Image ranges
#     x1, x2 = max(0, x), min(big_im.shape[0], x + small_im.shape[0])
#     y1, y2 = max(0, y), min(big_im.shape[1], y + small_im.shape[1])
#
#     # Overlay ranges
#     x1o, x2o = max(0, -x), min(small_im.shape[0], big_im.shape[0] - x)
#     y1o, y2o = max(0, -y), min(small_im.shape[1], big_im.shape[1] - y)
#
#     # Exit if nothing to do
#     if y1 >= y2 or x1 >= x2 or y1o >= y2o or x1o >= x2o:
#         return big_im
#
#     if mask is None:
#         big_im[x1:x2, y1:y2] += small_im[x1o:x2o, y1o:y2o]
#     else:
#         mask = mask[x1o:x2o, y1o:y2o].astype(bool)
#         big_im[x1:x2, y1:y2][mask] += small_im[x1o:x2o, y1o:y2o][mask]
#
#     return big_im
#

def overlay_image(small_im, big_im, coords, mask=None, sub_pixel=False):
    """
    Overlay small_im on top of big_im at the position specified by coords

    ``small_im`` will be centred at ``coords``

    Adapted from:
    ``https://stackoverflow.com/questions/14063070/overlay-a-smaller-image-on-a-larger-image-python-opencv``

    """

    # TODO - Add in a catch for sub-pixel shifts
    if sub_pixel:
        raise NotImplementedError

    y, x = np.array(coords, dtype=int)[::-1] - np.array(small_im.shape[-2:]) // 2

    # Image ranges
    x1, x2 = max(0, x), min(big_im.shape[-1], x + small_im.shape[-1])
    y1, y2 = max(0, y), min(big_im.shape[-2], y + small_im.shape[-2])

    # Overlay ranges
    x1o, x2o = max(0, -x), min(small_im.shape[-1], big_im.shape[-1] - x)
    y1o, y2o = max(0, -y), min(small_im.shape[-2], big_im.shape[-2] - y)

    # Exit if nothing to do
    if y1 >= y2 or x1 >= x2 or y1o >= y2o or x1o >= x2o:
        return big_im

    if small_im.ndim == 2 and big_im.ndim == 2:
        small_im_3 = small_im[None, :, :]
        big_im_3 = big_im[None, :, :]
    elif small_im.ndim == 3 and big_im.ndim == 3:
        small_im_3 = small_im
        big_im_3 = big_im
    else:
        raise ValueError(f"Dimensions mismatch between big_im and small_im: "
                         f"{big_im.ndim} : {small_im.ndim}")

    if mask is None:
        big_im_3[:, y1:y2, x1:x2] = big_im_3[:, y1:y2, x1:x2] + \
                                    small_im_3[:, y1o:y2o, x1o:x2o]
    else:
        mask = mask[None, y1o:y2o, x1o:x2o] * np.ones(small_im_3.shape[-3])
        mask = mask.astype(bool)
        big_im_3[:, y1:y2, x1:x2][mask] = big_im_3[:, y1:y2, x1:x2][mask] + \
                                          small_im_3[:, y1o:y2o, x1o:x2o][mask]

    return big_im


def rescale_imagehdu(imagehdu, pixel_scale, wcs_suffix="", conserve_flux=True,
                     spline_order=1):
    """
    Scales the .data array by the ratio of pixel_scale [deg] and CDELTn

    ``pixel_scale`` should NOT be passed as a Quantity!

    Parameters
    ----------
    imagehdu : fits.ImageHDU
    pixel_scale : float
        [deg] NOT to be passed as a Quantity
    wcs_suffix : str

    conserve_flux : bool

    spline_order : int
        [1..5] Order of the spline interpolation used by
        ``scipy.ndimage.rotate``

    Returns
    -------
    imagehdu : fits.ImageHDU

    """
    wcs_suffix = wcs_suffix
    s0 = wcs_suffix[0] if len(wcs_suffix) > 0 else ""
    cdelt1 = imagehdu.header["CDELT1"+s0]
    cdelt2 = imagehdu.header["CDELT2"+s0]

    zoom1 = np.abs(cdelt1 / pixel_scale)
    zoom2 = np.abs(cdelt2 / pixel_scale)  # making sure that zoom1,zoom2 are positive

    zoom_tuple = (zoom2, zoom1)
    if imagehdu.data.ndim == 3:
        zoom_tuple = (1, ) + zoom_tuple

    if zoom1 != 1 or zoom2 != 1:
        sum_orig = np.sum(imagehdu.data)
        new_im = ndi.zoom(imagehdu.data, zoom_tuple, order=spline_order)

        if conserve_flux:
            new_im = np.nan_to_num(new_im, copy=False)
            sum_new = np.sum(new_im)
            if sum_new != 0:
                new_im *= sum_orig / sum_new

        imagehdu.data = new_im

        for ii in range(max(1, len(wcs_suffix))):
            si = wcs_suffix[ii] if len(wcs_suffix) > 0 else ""
            imagehdu.header["CRPIX1"+si] *= zoom1
            imagehdu.header["CRPIX2"+si] *= zoom2
            imagehdu.header["CDELT1"+si] = pixel_scale
            imagehdu.header["CDELT2"+si] = pixel_scale
            imagehdu.header["CUNIT1"+si] = "mm" if si == 'D' else "deg"
            imagehdu.header["CUNIT2"+si] = "mm" if si == 'D' else "deg"

    return imagehdu


def reorient_imagehdu(imagehdu, wcs_suffix="", conserve_flux=True,
                      spline_order=1):
    """
    Applies an affine transformation to the image, as given in its header

    Parameters
    ----------
    imagehdu : fits.ImageHDU

    wcs_suffix : str

    conserve_flux : bool

    spline_order : int
        [1..5] Order of the spline interpolation used by
        ``scipy.ndimage.rotate``

    Returns
    -------
    imagehdu : fits.ImageHDU

    """

    s = wcs_suffix

    hdr = imagehdu.header
    pc_keys = ["PC1_1", "PC1_2", "PC2_1", "PC2_2"]
    if all(key+s in hdr for key in pc_keys) and imagehdu.data is not None:
        xscen, yscen = pix2val(hdr, hdr["NAXIS1"] / 2., hdr["NAXIS2"] / 2., s)
        hdr["CRVAL1" + s] = xscen
        hdr["CRVAL2" + s] = yscen

        mat = np.array([[hdr["PC1_1" + s], hdr["PC1_2" + s], 0],
                        [hdr["PC2_1" + s], hdr["PC2_2" + s], 0],
                        [0,                0,                1]])
        if imagehdu.data.ndim == 2:
            mat = mat[:2, :2]

        new_im = affine_map(imagehdu.data, matrix=mat, reshape=True,
                            spline_order=spline_order)
        # new_im = ndi.rotate(imagehdu.data, angle, reshape=True, order=spline_order)

        if conserve_flux:
            new_im = np.nan_to_num(new_im, copy=False)
            new_im *= np.sum(imagehdu.data) / np.sum(new_im)

        imagehdu.data = new_im
        hdr["CRPIX1"+s] = hdr["NAXIS1"] / 2.
        hdr["CRPIX2"+s] = hdr["NAXIS2"] / 2.
        for card in ["PC1_1"+s, "PC1_2"+s, "PC2_1"+s, "PC2_2"+s]:
            hdr.remove(card)
        imagehdu.header = hdr

    elif any(["PC1_1" in key for key in imagehdu.header]):
        logging.warning("PC Keywords were found, but not used due to different "
                      "wcs_suffix given: {} \n {}"
                      "".format(wcs_suffix, dict(imagehdu.header)))

    return imagehdu


def affine_map(input, matrix=None, rotation_angle=0, shear_angle=0,
               scale_factor=None, reshape=True, spline_order=3):
    """
    Applies an affine transformation matrix to an image around its centre

    Similar functionality to ``scipy.ndimage.rotate`` but for the
    ``affine_transformation`` function

    Either a 2x2 affine transformation matrix can be supplied, or the rotation,
    shear, and scaling values which are the basis of an affine transformation.


    Parameters
    ----------
    input : array_like
        The input array
    matrix : ndarray, optional
        A 2x2 affine transformation matrix
    rotation_angle : float, optional
        [deg] If matrix==None, a rotation matrix is built from this angle
    shear_angle : float, optional
        [deg] If matrix==None, a y-axis shear matrix is built from this angle
    scale_factor : list, array
        [mx, my] If matrix==None, a scaling matrix is built from this list
    reshape : bool, optional
        If True, the array is re-sized to contain the whole transformed image
    spline_order : int, optional
        Default is 3. Spline interpolation order

    Returns
    -------
    output : array-like
        The new mapping of the image

    """

    if matrix is None:
        d2r = np.pi / 180.
        c = np.cos(rotation_angle * d2r) if rotation_angle != 0. else 1.
        s = np.sin(rotation_angle * d2r) if rotation_angle != 0. else 0.
        t = np.tan(shear_angle * d2r) if shear_angle != 0. else 0.
        sx = scale_factor[0] if scale_factor is not None else 1.
        sy = scale_factor[1] if scale_factor is not None else 1.
        mat = np.array([[c, s], [-s, c]]) @ \
              np.array([[1, t], [0, 1]]) @ \
              np.array([[sx, 0], [0, sy]])
    else:
        mat = np.array(matrix)

    h, w = np.array(input.shape)[-2:]
    if reshape:
        # Find the new corner coordinates using the py3.5+ matmul operator
        x, y = mat[:2, :2] @ np.array([[0, w, w, 0], [0, 0, h, h]])
        out = np.ceil(np.array([y.max() - y.min(),
                                x.max() - x.min()])).astype(int)
    else:
        out = np.array([h, w])

    c_out = 0.5 * out
    c_in = 0.5 * np.array([h, w])
    offset = c_in - c_out.dot(mat[:2, :2])

    if mat.shape[0] == 3:
        offset = np.r_[[0], offset]
        out = np.r_[input.shape[0], out]

    output = ndi.affine_transform(input, np.rot90(mat, 2),
                                  output_shape=out, offset=offset,
                                  order=spline_order)

    return output


def add_imagehdu_to_imagehdu(image_hdu, canvas_hdu, spline_order=1,
                             wcs_suffix="", conserve_flux=True):
    """
    Re-project one ``fits.ImageHDU`` onto another ``fits.ImageHDU``

    ..assumption:: of equal grid coordinate lengths

    Parameters
    ----------
    image_hdu : fits.ImageHDU
        The ``ImageHDU`` which will be reprojected onto `canvas_hdu`

    canvas_hdu : fits.ImageHDU
        The ``ImageHDU`` onto which the image_hdu should be projected.
        This must include a valid WCS

    spline_order : int, optional
        Default is 1. The order of the spline interpolator used by the
        ``scipy.ndimage`` functions

    wcs_suffix : str
        To determine which WCS to use. "" for sky HDUs and "D" for
        ImagePlane HDUs

    conserve_flux : bool
        Default is True. Used when zooming and rotating to keep flux constant.

    Returns
    -------
    canvas_hdu : fits.ImageHDU

    """

    # .. todo: Add a catch for projecting a large image onto a small canvas

    if isinstance(image_hdu.data, u.Quantity):
        image_hdu.data = image_hdu.data.value
    pixel_scale = float(canvas_hdu.header["CDELT1"+wcs_suffix])

    new_hdu = rescale_imagehdu(image_hdu, pixel_scale=pixel_scale,
                               wcs_suffix=wcs_suffix, spline_order=spline_order,
                               conserve_flux=conserve_flux)
    new_hdu = reorient_imagehdu(new_hdu,
                                wcs_suffix=wcs_suffix, spline_order=spline_order,
                                conserve_flux=conserve_flux)

    xcen_im = (new_hdu.header["NAXIS1"] - 1) / 2 #// 2
    ycen_im = (new_hdu.header["NAXIS2"] - 1) / 2 #// 2

    xsky0, ysky0 = pix2val(new_hdu.header, xcen_im, ycen_im, wcs_suffix)
    xpix0, ypix0 = val2pix(canvas_hdu.header, xsky0, ysky0, wcs_suffix)

    # again, I need to add this transpose operation - WHY????
    # Image plane tests need the transpose operation, but FOV broadcast tests don't. Weird
    canvas_hdu.data = overlay_image(new_hdu.data, canvas_hdu.data,
                                    coords=(xpix0+1, ypix0+1))
                                    #coords=(xpix0, ypix0))

    return canvas_hdu


def pix2val(header, x, y, wcs_suffix=""):
    """
    Returns the real coordinates [deg, mm] for coordinates from a Header WCS

    Parameters
    ----------
    header : fits.Header
    x, y : float, list, array
    wcs_suffix : str

    Returns
    -------
    a, b : float, array
        [deg, mm] Real coordinates as given by the Header WCS

    """

    s = wcs_suffix
    if "PC1_1"+s in header:
        pc11 = header["PC1_1"+s]
        pc12 = header["PC1_2"+s]
        pc21 = header["PC2_1"+s]
        pc22 = header["PC2_2"+s]
    else:
        pc11, pc12, pc21, pc22 = 1, 0, 0, 1

    da = header["CDELT1"+s]
    db = header["CDELT2"+s]
    x0 = header["CRPIX1"+s]
    y0 = header["CRPIX2"+s]
    a0 = header["CRVAL1"+s]
    b0 = header["CRVAL2"+s]

    a = a0 + da * ((x - x0) * pc11 + (y - y0) * pc12)
    b = b0 + db * ((x - x0) * pc21 + (y - y0) * pc22)

    return a, b


def val2pix(header, a, b, wcs_suffix=""):
    """
    Returns the pixel coordinates for real coordinates [deg, mm] from a WCS

    Parameters
    ----------
    header : fits.Header
    a, b : float, list, array
        [deg, mm]

    Returns
    -------
    x, y : float, array
        [pixel] Pixel coordinates as given by the Header WCS

    """

    s = wcs_suffix

    if isinstance(header, fits.ImageHDU):
        header = header.header

    if "PC1_1"+s in header:
        pc11 = header["PC1_1"+s]
        pc12 = header["PC1_2"+s]
        pc21 = header["PC2_1"+s]
        pc22 = header["PC2_2"+s]
    else:
        pc11, pc12, pc21, pc22 = 1, 0, 0, 1

    da = float(header["CDELT1"+s])
    db = float(header["CDELT2"+s])
    x0 = float(header["CRPIX1"+s])
    y0 = float(header["CRPIX2"+s])
    a0 = float(header["CRVAL1"+s])
    b0 = float(header["CRVAL2"+s])

    x = x0 + 1. / da * ((a - a0) * pc11 - (b - b0) * pc21)
    y = y0 + 1. / db * ((a - a0) * pc12 + (b - b0) * pc22)

    return x, y


def calc_footprint(header, wcs_suffix=""):
    """
    Returns the sky/detector positions [deg/mm] of the corners of a header WCS

    The positions returned correspond to the corners of the header's
    image array, in this order::

        (ra, dec) = (0,0), (w, 0), (w, h), (0, h)
        (x, y) = (0,0), (w, 0), (w, h), (0, h)

    where ``w``, ``h`` are equal to NAXIS1 and NAXIS2 from the header.

    Parameters
    ----------
    header : fits.Header

    wcs_suffix : str
        Letter suffix for the WCS keywords, e.g. CDELT1D for image-plane coords

    Returns
    -------
    x, y : arrays of floats
        [deg or mm] x are the coordinates for pixels [0, w, w, 0]
        [deg or mm] y are the coordinates for pixels [0, 0, h, h]

    """

    if isinstance(header, fits.ImageHDU):
        header = header.header

    w, h = header["NAXIS1"], header["NAXIS2"]
    x0 = np.array([0, w, w, 0])
    y0 = np.array([0, 0, h, h])

    x1, y1 = pix2val(header, x0, y0, wcs_suffix)

    return x1, y1


def split_header(hdr, chunk_size, wcs_suffix=""):
    """
    Splits a header into many smaller parts of the chunk_size

    Parameters
    ----------
    hdr
    chunk_size
    wcs_suffix

    Returns
    -------

    """
    # ..todo:: test that this works
    s = wcs_suffix
    naxis1, naxis2 = hdr["NAXIS1"+s], hdr["NAXIS2"+s]
    x0_pix, y0_pix = hdr["CRPIX1"+s], hdr["CRPIX2"+s]       # pix
    x0_sky, y0_sky = hdr["CRVAL1"+s], hdr["CRVAL2"+s]       # deg
    x_delt, y_delt = hdr["CDELT1"+s], hdr["CDELT2"+s]       # deg / pix

    hdr_list = []
    for x1_pix in range(0, naxis1, chunk_size):
        for y1_pix in range(0, naxis2, chunk_size):
            x1_sky = x0_sky + (x1_pix - x0_pix) * x_delt
            y1_sky = y0_sky + (y1_pix - y0_pix) * y_delt
            x2_sky = x1_sky + x_delt * min(chunk_size, naxis1 - x1_pix)
            y2_sky = y1_sky + y_delt * min(chunk_size, naxis2 - y1_pix)

            hdr_sky = header_from_list_of_xy([x1_sky, x2_sky], [y1_sky, y2_sky],
                                             pixel_scale=x_delt, wcs_suffix=s)
            hdr_list += [hdr_sky]

    return hdr_list



# takes a package name as input
# reads in all yaml files within the package
# creates rst descriptions for each optical element
# shows the system throughput
# list all the available modes
# list the dependencies
# produces all the plots
# produces all the rst files
# produces a top-level rst / latex file that references all the generated ones
# can read in preambles / other package specific descriptions


class ReportGenerator:
    def __init__(self, pkg_name):
        self.pkg_name = pkg_name

    def rst_default_yaml(self):
        pass






import yaml
from copy import deepcopy
import datetime
import numpy as np
from astropy.io import fits
from astropy import units as u
from astropy.table import Table
from . import Effect
from ..utils import check_keys, from_currsys, find_file


class ExtraFitsKeywords(Effect):
    """
    Extra FITS header keywords to be added to the pipeline FITS files

    These keywords are ONLY for keywords that should be MANUALLY ADDED to the
    headers after a simulation is read-out by the detector.

    Simulation parameters (Effect kwargs values, etc) will be added automatically
    by ScopeSim in a different function, but following this format.

    The dictionaries should be split into different HIERARCH lists, e.g.:

    - HIERARCH ESO
      For ESO specific keywords
    - HIERARCH SIM
      For ScopeSim specific keywords, like simulation parameters
    - HIERARCH MIC
      For MICADO specific keywords, (unsure what these would be yet)

    More HIERARCH style keywords can also be added as needed for other use-cases

    Parameters
    ----------
    filename : str, optional
        Name of a .yaml nested dictionary file. See below for examples

    yaml_string : str, optional
        A triple-" string containing the contents of a yaml file

    header_dict : nested dicts, optional
        A series of nested python dictionaries following the format of the
        examples below. This keyword allows these dicts to be definied directly
        in the Effect yaml file, rather than in a seperate header keywords file.

    Examples
    --------
    Specifying the extra FITS keywords directly in the .yaml file where the
    Effect objects are described.
    ::

        name: extra_fits_header_entries
        class: ExtraFitsKeywords
        kwargs:
          header_dict:
            - ext_type: PrimaryHDU
              keywords:
                HIERARCH:
                  ESO:
                    ATM:
                      TEMPERAT: -5

    The contents of ``header_dict`` can also be abstracted away into a seperate
    file, e.g. ``extra_FITS_keys.yaml``. The file format is described below in
    detail below.
    ::

        name: extra_fits_header_entries
        class: ExtraFitsKeywords
        kwargs:
          filename: extra_FITS_keys.yaml

    The Effect can be added directly in an iPython session.
    ::

        >>> hdr_dic = {"ext_type": "PrimaryHDU",
                       "keywords":
                           {"HIERARCH":
                               {"SIM":
                                   {"hello": world}
                               }
                           }
                       }
        >>> extra_keys = ExtraFitsKeywords(header_dict=hdr_dic)
        >>> optical_train.optics_manager.add_effect(extra_keys)


    Yaml file format
    ----------------
    This document is a yaml document.
    Hence all new keywords should be specified in the form of yaml nested
    dictionaries.
    As each ``astropy.HDUList`` contains one or more extensions, the inital level is
    reserved for a list of keyword groups.
    For example::

        - ext_type: PrimaryHDU
          keywords:
            HIERARCH:
              ESO:
                ATM:
                  TEMPERAT: -5

        - ext_number: [1, 2]
          keywords:
            HIERARCH:
              ESO:
                DET:
                  DIT: [5, '[s] exposure length']   # example of adding a comment
            EXTNAME: "DET++.DATA"                   # example of extension specific qualifier

    The keywords can be added to one or more extensions, based on one of the
    following ``ext_`` qualifiers: ``ext_name``, ``ext_number``, ``ext_type``

    Each of these ``ext_`` qualifiers can be a ``str`` or a ``list``.
    For a list, ScopeSim will add the keywords to all extensions matching the
    specified type/name/number

    The number of the extension can be used in a value by using the "++" string.
    That is, keyword values with "++" with have the extension number inserted
    where the "++" is.

    The above example (``EXTNAME: "DET++.DATA"``) will result in the following
    keyword added only to extensions 1 and 2:

    - PrimaryHDU (ext 0)::

          header['HIERARCH ESO ATM TEMPERAT'] = -5

    - Extension 1 (regardless of type)::

          header['HIERARCH ESO DET DIT'] = (5, '[s] exposure length')
          header['EXTNAME'] = "DET1.DATA"

    - Extension 2 (regardless of type)::

          header['HIERARCH ESO DET DIT'] = (5, '[s] exposure length')
          header['EXTNAME'] = "DET2.DATA"

    Resolved and un-resolved keywords
    ---------------------------------
    ScopeSim uses bang-strings to resolve global parameters.
    E.g: ``from_currsys('!ATMO.temperature')`` will resolve to a float
    These bang-strings will be resolved automatically in the ``keywords``
    dictionary section.

    If the keywords bang-string should instead remain unresolved and the string
    added verbatim to the header, we use the ``unresolved_keywords`` dictionary
    section.

    Additionally, new functionality will be added to ScopeSim to resolve the
    kwargs/meta parameters of Effect objects.
    The format for this will be to use a new type: the hash-string.
    This will have this format::

        #<optical_element_name>.<effect_name>.<kwarg_name>

    For example, the temperature of the MICADO detector array can be accessed
    by::

        '#MICADO_DET.full_detector_array.temperature'

    In the context of the yaml file this would look like::

        - ext_type: PrimaryHDU
          keywords:
            HIERARCH:
              ESO:
                DET
                  TEMPERAT: '#MICADO_DET.full_detector_array.temperature'

    Obviously some though needs to be put into how exactly we list the
    simulation parameters in a coherent manner.
    But this is 'Zukunftsmusik'.
    For now we really just want an interface that can add the ESO header
    keywords, which can also be expanded in the future for our own purposes.

    Below is an example of some extra keywords for MICADO headers::

        - ext_type: PrimaryHDU
          keywords:
            HIERARCH:
              ESO:
                ATM:
                  TEMPERAT: '!ATMO.temperature'   # will be resolved via from_currsys
                  PWV: '!ATMO.pwv'
                  SEEING: 1.2
                DAR:
                  VALUE: '#<effect_name>.<meta_name>'   # will be resolved via effects
                DPR:
                  TYPE: 'some_type'
              SIM:
                random_simulation_keyword: some_value
              MIC:
                micado_specific: ['keyword', 'keyword comment']

          unresolved_keywords:
            HIERARCH:
              ESO:
                ATM:
                  TEMPERAT: '!ATMO.temperature'   # will be left as a string

        - ext_type: ImageHDU
          keywords:
            HIERARCH:
              SIM:
                hello: world
                hallo: welt
                grias_di: woed
                zdrasviute: mir
                salud: el mundo

    """
    def __init__(self, **kwargs):
        # don't pass kwargs, as DataContainer can't handle yaml files
        super(ExtraFitsKeywords, self).__init__()
        params = {"name": "extra_fits_keywords",
                  "description": "Extra FITS headers",
                  "z_order": [999],
                  "header_dict": None,
                  "filename": None,
                  "yaml_string": None,
                  }
        self.meta.update(params)
        self.meta.update(kwargs)

        tmp_dicts = []
        if self.meta["filename"] is not None:
            yaml_file = find_file(self.meta["filename"])
            with open(yaml_file) as f:
                # possible multiple yaml docs in a file
                # --> returns list even for a single doc
                tmp_dicts += [dic for dic in yaml.full_load_all(f)]

        if self.meta["yaml_string"] is not None:
            yml = self.meta["yaml_string"]
            tmp_dicts += [dic for dic in yaml.full_load_all(yml)]

        if self.meta["header_dict"] is not None:
            if not isinstance(self.meta["header_dict"], list):
                tmp_dicts += [self.meta["header_dict"]]
            else:
                tmp_dicts += self.meta["header_dict"]

        self.dict_list = []
        for dic in tmp_dicts:
            # format says yaml file contains list of dicts
            if isinstance(dic, list):
                self.dict_list += dic
            # catch case where user forgets the list
            elif isinstance(dic, dict):
                self.dict_list += [dic]

    def apply_to(self, hdul, **kwargs):
        """
        Adds extra fits keywords from a yaml file including !,#-stings

        Parameters
        ----------
        optical_train : scopesim.OpticalTrain, optional
            Used to resolve #-strings

        """
        opt_train = kwargs.get("optical_train")
        if isinstance(hdul, fits.HDUList):
            for dic in self.dict_list:
                resolved = flatten_dict(dic.get("keywords", {}), resolve=True,
                                        optics_manager=opt_train)
                unresolved = flatten_dict(dic.get("unresolved_keywords", {}))
                exts = get_relevant_extensions(dic, hdul)
                for i in exts:
                    resolved_with_counters = {
                        k: v.replace("++", str(i)) if isinstance(v, str) else v
                        for k, v in resolved.items()
                    }
                    hdul[i].header.update(resolved_with_counters)
                    hdul[i].header.update(unresolved)

        return hdul


def get_relevant_extensions(dic, hdul):
    exts = []
    if dic.get("ext_name") is not None:
        exts += [i for i, hdu in enumerate(hdul)
                 if hdu.header["EXTNAME"] == dic["ext_name"]]
    elif dic.get("ext_number") is not None:
        ext_n = np.array(dic["ext_number"])
        exts += list(ext_n[ext_n<len(hdul)])
    elif dic.get("ext_type") is not None:
        if isinstance(dic["ext_type"], list):
            ext_type_list = dic["ext_type"]
        else:
            ext_type_list = [dic["ext_type"]]
        cls = tuple([getattr(fits, cls_str) for cls_str in ext_type_list])
        exts += [i for i, hdu in enumerate(hdul) if isinstance(hdu, cls)]

    return exts


def flatten_dict(dic, base_key="", flat_dict=None,
                 resolve=False, optics_manager=None):
    """
    Flattens nested yaml dictionaries into a single level dictionary

    Parameters
    ----------
    dic : dict
    base_key : str
    flat_dict : dict, optional
        Top-level dictionary for recursive calls
    resolve : bool
        If True, resolves !-str via from_currsys and #-str via optics_manager
    optics_manager : scopesim.OpticsManager
        Required for resolving #-strings

    Returns
    -------
    flat_dict : dict

    """
    if flat_dict is None:
        flat_dict = {}
    for key, val in dic.items():
        flat_key = base_key + f"{key} "
        if isinstance(val, dict):
            flatten_dict(val, flat_key, flat_dict, resolve, optics_manager)
        else:
            flat_key = flat_key[:-1]

            # catch any value+comments lists
            comment = ""
            if isinstance(val, list) and len(val) == 2 and isinstance(val[1], str):
                value, comment = val
            else:
                value = deepcopy(val)

            # resolve any bang or hash strings
            if resolve and isinstance(value, str) and len(value) > 0:
                if value[0] == "!":
                    value = from_currsys(value)
                elif value[0] == "#":
                    if optics_manager is None:
                        raise ValueError("An OpticsManager object must be "
                                         "passed in order to resolve #-strings")
                    value = optics_manager[value]

            if isinstance(value, u.Quantity):
                comment = f"[{str(value.unit)}] " + comment
                value = value.value

            if isinstance(value, (list, np.ndarray)):
                value = f"{value.__class__.__name__}:{str(list(value))}"
                max_len = 80 - len(flat_key)
                if len(value) > max_len:
                    value = value[:max_len-4] + " ..."

            if isinstance(value, (datetime.time, datetime.date, datetime.datetime)):
                value = value.isoformat()

            # Add the flattened KEYWORD = (value, comment) to the header dict
            if len(comment) > 0:
                flat_dict[flat_key] = (value, str(comment))
            else:
                flat_dict[flat_key] = value

    return flat_dict


class EffectsMetaKeywords(ExtraFitsKeywords):
    """
    Adds meta dictionary info from all Effects to the FITS headers

    Parameters
    ----------
    ext_number : int, list of ints, optional
        Default 0. The numbers of the extensions to which the header keywords
        should be added

    add_excluded_effects : bool, optional
        Default False. Add meta dict for effects with ``<effect>.include=False``

    keyword_prefix : str, optional
        Default "HIERARCH SIM". Custom FITS header keyword prefix. Effect meta
        dict entries will appear in the header as:
        ``<keyword_prefix> EFFn <key> : <value>``

    Examples
    --------
    Yaml file entry:
    ::

        name: effect_dumper
        class: EffectsMetaKeywords
        description: adds all effects meta dict entries to the FITS header
        kwargs:
          ext_number: [0, 1]
          add_excluded_effects: False
          keyword_prefix: HIERARCH SIM

    """
    def __init__(self, **kwargs):
        super(ExtraFitsKeywords, self).__init__()
        params = {"name": "effects_fits_keywords",
                  "description": "Effect Meta FITS headers",
                  "z_order": [998],
                  "ext_number": [0],
                  "add_excluded_effects": False,
                  "keyword_prefix": "HIERARCH SIM"}
        self.meta.update(params)
        self.meta.update(kwargs)

    def apply_to(self, hdul, **kwargs):
        opt_train = kwargs.get("optical_train")
        if isinstance(hdul, fits.HDUList) and opt_train is not None:
            # todo: use a different way of getting all the effect names
            # opt.effects returns the __repr__, not the original name
            for i, eff_name in enumerate(opt_train.effects["name"]):
                # Check for spaces
                if " " in eff_name:
                    # E.g. 'filter_wheel_1 : [open]'
                    assert "wheel" in eff_name, \
                        f"Unknown effect name with space: {eff_name}"
                    eff_name = eff_name.split()[0]

                # get a resolved meta dict from the effect
                eff_meta = deepcopy(opt_train[f"#{eff_name}.!"])

                if self.meta["add_excluded_effects"] and not eff_meta["include"]:
                    continue

                keys = list(eff_meta.keys())
                for key in keys:
                    value = eff_meta[key]
                    if key in ["history", "notes", "changes"]:
                        eff_meta.pop(key)
                    if isinstance(value, Table):
                        eff_meta[key] = f"Table object of length: {len(value)}"

                # add effect under the EFFn keyword
                prefix = self.meta['keyword_prefix']
                class_name = opt_train[eff_name].__class__.__name__
                self.dict_list = [{"ext_number": self.meta["ext_number"],
                                   "keywords": {
                                       f"{prefix} EFF{i} class": [class_name, "ScopeSim class name"],
                                       f"{prefix} EFF{i}": eff_meta}
                                   }]
                super_apply_to = super(EffectsMetaKeywords, self).apply_to
                hdul = super_apply_to(hdul=hdul, optical_train=opt_train)

        return hdul


class SourceDescriptionFitsKeywords(ExtraFitsKeywords):
    """
    Adds parameters from all Source fields to the FITS headers

    Parameters
    ----------
    ext_number : int, list of ints, optional
        Default 0. The numbers of the extensions to which the header keywords
        should be added

    keyword_prefix : str, optional
        Default "HIERARCH SIM". Custom FITS header keyword prefix. Effect meta
        dict entries will appear in the header as:
        ``<keyword_prefix> SRCn <key> : <value>``

    Examples
    --------
    Yaml file entry:
    ::

        name: source_descriptor
        class: SourceDescriptionFitsKeywords
        description: adds info from all Source fields to the FITS header
        kwargs:
          ext_number: [0]
          keyword_prefix: HIERARCH SIM

    """
    def __init__(self, **kwargs):
        super(ExtraFitsKeywords, self).__init__()
        params = {"name": "source_fits_keywords",
                  "description": "Source description FITS headers",
                  "z_order": [997],
                  "ext_number": [0],
                  "keyword_prefix": "HIERARCH SIM"}
        self.meta.update(params)
        self.meta.update(kwargs)

    def apply_to(self, hdul, **kwargs):
        opt_train = kwargs.get("optical_train")
        if isinstance(hdul, fits.HDUList) and opt_train is not None:
            src = opt_train._last_source
            src_dicts = []
            if src is not None:
                prefix = self.meta['keyword_prefix']
                for i, field in enumerate(src.fields):

                    src_class = field.__class__.__name__
                    src_dic = deepcopy(src._meta_dicts[i])
                    if isinstance(field, fits.ImageHDU):
                        hdr = field.header
                        for key in hdr:
                            src_dic = {key: [hdr[key], hdr.comments[key]]}

                    elif isinstance(field, Table):
                        src_dic.update(field.meta)
                        src_dic["length"] = len(field)
                        for j, name in enumerate(field.colnames):
                            src_dic[f"col{j}_name"] = name
                            src_dic[f"col{j}_unit"] = str(field[name].unit)

                    self.dict_list = [{"ext_number": self.meta["ext_number"],
                                       "keywords": {
                                           f"{prefix} SRC{i} class": src_class,
                                           f"{prefix} SRC{i}": src_dic}
                                       }]
                    super_apply_to = super(SourceDescriptionFitsKeywords, self).apply_to
                    hdul = super_apply_to(hdul=hdul, optical_train=opt_train)

            # catch the function call
            for hdu in hdul:
                for key in hdu.header:
                    if "function_call" in key:
                        hdu.header[f"FN{key.split()[1]}"] = hdu.header.pop(key)

        return hdul


class SimulationConfigFitsKeywords(ExtraFitsKeywords):
    """
    Adds parameters from all config dictionaries to the FITS headers

    Parameters
    ----------
    ext_number : int, list of ints, optional
        Default 0. The numbers of the extensions to which the header keywords
        should be added

    resolve : bool
        Default True. If True, all !-strings and #-strings are resolved via
        ``from_currsys`` before being add to the header. If False, the
        unaltered !-strings or #-strings are added to the header.

    keyword_prefix : str, optional
        Default "HIERARCH SIM". Custom FITS header keyword prefix. Effect meta
        dict entries will appear in the header as:
        ``<keyword_prefix> SRCn <key> : <value>``

    Examples
    --------
    Yaml file entry:
    ::

        name: source_descriptor
        class: SimulationConfigFitsKeywords
        description: adds info from all config dicts to the FITS header
        kwargs:
          ext_number: [0]
          resolve: False
          keyword_prefix: HIERARCH SIM

    """
    def __init__(self, **kwargs):
        super(ExtraFitsKeywords, self).__init__()
        params = {"name": "simulation_fits_keywords",
                  "description": "Simulation Config FITS headers",
                  "z_order": [996],
                  "ext_number": [0],
                  "resolve": True,
                  "keyword_prefix": "HIERARCH SIM"}
        self.meta.update(params)
        self.meta.update(kwargs)

    def apply_to(self, hdul, **kwargs):
        opt_train = kwargs.get("optical_train")
        if isinstance(hdul, fits.HDUList) and opt_train is not None:
            cmds = opt_train.cmds.cmds.dic
            sim_prefix = self.meta['keyword_prefix']
            resolve_prefix = "unresolved_" if not self.meta["resolve"] else ""
            # needed for the super().apply_to method
            self.dict_list = [{"ext_number": self.meta["ext_number"],
                               f"{resolve_prefix}keywords": {
                                   f"{sim_prefix} CONFIG": cmds}
                               }]
            super_apply_to = super(SimulationConfigFitsKeywords, self).apply_to
            hdul = super_apply_to(hdul=hdul, optical_train=opt_train)

        return hdul




import numpy as np

from astropy import units as u

from synphot import SourceSpectrum, SpectralElement
from synphot.models import Empirical1D

from .. import rc
from .. import utils
from ..utils import quantify
from ..base_classes import SourceBase, ImagePlaneBase, FieldOfViewBase
from ..optics.radiometry import RadiometryTable
from .effects import Effect
from .ter_curves import TERCurve


class SurfaceList(Effect):
    """
    A Effect object containing a list of all surfaces in an optical element

    This calls is essentially a wrapper for the RadiometryTable class, which
    contains all the functionality for creating and combining multiple optical
    surfaces


    Parameters
    ----------
    filename : str
        Path to file containing
    table : astropy.Table

    array_dict : dict


    Input format
    ------------

    """
    def __init__(self, **kwargs):
        super(SurfaceList, self).__init__(**kwargs)
        self.meta["z_order"] = [20, 120]
        self.meta["minimum_throughput"] = "!SIM.spectral.minimum_throughput"
        self.meta["wave_min"] = "!SIM.spectral.wave_min"
        self.meta["wave_max"] = "!SIM.spectral.wave_max"
        self.meta.update(kwargs)

        self.radiometry_table = RadiometryTable()
        self.radiometry_table.meta.update(self.meta)
        self._emission = None
        self._throughput = None

        data = self.get_data()
        if data is not None:
            self.radiometry_table.add_surface_list(data)

    def apply_to(self, obj, **kwargs):
        """
        obj == SourceBase - applies throughput
        obj == ImagePlaneBase - applies emission if Imager
        obj == FieldOfViewBase - applies emission if Spectrograph

        """
        if isinstance(obj, SourceBase) and not self.is_empty:
            self.meta = utils.from_currsys(self.meta)
            for ii in range(len(obj.spectra)):
                spec = obj.spectra[ii]
                wave_val = spec.waveset.value
                wave_unit = spec.waveset.unit  # angstrom
                wave_min = quantify(self.meta["wave_min"], u.um).to(u.AA)
                wave_max = quantify(self.meta["wave_max"], u.um).to(u.AA)
                mask = (wave_val > wave_min.value) * (wave_val < wave_max.value)

                wave = ([wave_min.value] +
                        list(wave_val[mask]) +
                        [wave_max.value]) * wave_unit
                thru = self.throughput(wave)
                flux = spec(wave)
                flux *= thru
                new_source = SourceSpectrum(Empirical1D, points=wave,
                                            lookup_table=flux)
                obj.spectra[ii] = new_source

        elif isinstance(obj, ImagePlaneBase) and not self.is_empty:
            # by calling use_area, the surface area is taken into account, but
            # the units are stuck in PHOTLAM for synphot
            emission = self.get_emission(use_area=True)  # --> PHOTLAM * area
            if emission is not None:
                wave = emission.waveset  # angstrom
                flux = emission(wave)    # PHOTLAM --> ph s-1 cm-2 AA-1 * cm2
                phs = (np.trapz(flux, wave) * u.cm**2).to(u.Unit("ph s-1"))
            else:
                phs = 0 * (u.ph / u.s)

            obj.hdu.data += phs.value

        elif isinstance(obj, FieldOfViewBase) and not self.is_empty:
            # ..todo:: Super hacky, FIX THIS!!
            emission = self.get_emission(use_area=True)  # --> PHOTLAM * area
            if emission is not None:
                wave_val = emission.waveset.value
                wave_unit = emission.waveset.unit   # angstrom
                wave_min = quantify(obj.meta["wave_min"], u.um).to(wave_unit)
                wave_max = quantify(obj.meta["wave_max"], u.um).to(wave_unit)
                mask = (wave_val > wave_min.value) * (wave_val < wave_max.value)

                wave = ([wave_min.value] + list(wave_val[mask]) +
                        [wave_max.value]) * wave_unit
                flux = emission(wave)    # PHOTLAM --> ph s-1 cm-2 AA-1 * cm2
                phs = (np.trapz(flux, wave) * u.cm**2).to(u.Unit("ph s-1"))
            else:
                phs = 0 * (u.ph / u.s)

            obj.hdu.data += phs.value

        return obj

    def fov_grid(self, which="waveset", **kwargs):
        if which == "waveset":
            self.meta.update(kwargs)
            self.meta = utils.from_currsys(self.meta)
            wave_min = utils.quantify(self.meta["wave_min"], u.um)
            wave_max = utils.quantify(self.meta["wave_max"], u.um)
            # ..todo:: add 1001 to default.yaml somewhere
            wave = np.linspace(wave_min, wave_max, 1001)
            throughput = self.throughput(wave)
            threshold = self.meta["minimum_throughput"]
            valid_waves = np.where(throughput >= threshold)[0]
            if len(valid_waves) > 0:
                wave_edges = [min(wave[valid_waves]), max(wave[valid_waves])]
            else:
                raise ValueError("No transmission found above the threshold {} "
                                 "in this wavelength range {}. Did you open "
                                 "the shutter?"
                                 "".format(self.meta["minimum_throughput"],
                                           [self.meta["wave_min"],
                                            self.meta["wave_max"]]))
        else:
            wave_edges = []

        return wave_edges

    def add_surface(self, surface, name, position=-1, add_to_table=True):
        if isinstance(surface, TERCurve):
            ter_meta = surface.meta
            surface = surface.surface
            surface.meta.update(ter_meta)
        self.radiometry_table.add_surface(surface, name, position, add_to_table)

    def add_surface_list(self, surface_list, prepend=False):
        if isinstance(surface_list, SurfaceList):
            surface_list = surface_list.radiometry_table.table
        elif isinstance(surface_list, RadiometryTable):
            surface_list = surface_list.table

        self.radiometry_table.add_surface_list(surface_list, prepend)

    def get_emission(self, **kwargs):
        etendue = rc.__currsys__["!TEL.etendue"]
        return self.radiometry_table.get_emission(etendue=etendue, **kwargs)

    def get_throughput(self, **kwargs):
        return self.radiometry_table.get_throughput(**kwargs)

    def collapse(self, waveset):
        throughput = self.radiometry_table.throughput(waveset)
        self._throughput = SpectralElement(Empirical1D, points=waveset,
                                           lookup_table=throughput)
        emission = self.radiometry_table.emission(waveset)
        self._emission = SourceSpectrum(Empirical1D, points=waveset,
                                        lookup_table=emission)

    @property
    def throughput(self):
        return self.get_throughput()

    @property
    def emission(self):
        return self.get_emission()

    @property
    def area(self):
        if not self.is_empty:
            tbl = self.radiometry_table.table
            outer_col = utils.real_colname("outer", tbl.colnames)
            inner_col = utils.real_colname("inner", tbl.colnames)
            outer = utils.quantity_from_table(outer_col, tbl, u.m)
            inner = utils.quantity_from_table(inner_col, tbl, u.m)
            scope_area = np.max(np.pi / 4 * (outer ** 2 - inner ** 2))
        else:
            scope_area = 0

        return scope_area

    @property
    def is_empty(self):
        return len(self.radiometry_table.table) == 0



import logging


class SystemDict(object):
    def __init__(self, new_dict=None):
        self.dic = {}
        if isinstance(new_dict, dict):
            self.update(new_dict)
        elif isinstance(new_dict, list):
            for entry in new_dict:
                self.update(entry)

    def update(self, new_dict):
        if isinstance(new_dict, dict) \
                and "alias" in new_dict \
                and "properties" in new_dict:
            alias = new_dict["alias"]
            if alias in self.dic:
                self.dic[alias] = recursive_update(self.dic[alias],
                                                   new_dict["properties"])
            else:
                self.dic[alias] = new_dict["properties"]
        else:
            "Catch any bang-string properties keys"
            to_pop = []
            for key in new_dict:
                if key[0] == "!":
                    self[key] = new_dict[key]
                    to_pop += [key]
            for key in to_pop:
                new_dict.pop(key)

            if len(new_dict) > 0:
                self.dic = recursive_update(self.dic, new_dict)

    def __getitem__(self, item):
        if isinstance(item, str) and item[0] == "!":
            item_chunks = item[1:].split(".")
            entry = self.dic
            for item in item_chunks:
                entry = entry[item]
            return entry
        else:
            return self.dic[item]

    def __setitem__(self, key, value):
        if isinstance(key, str) and key[0] == "!":
            key_chunks = key[1:].split(".")
            entry = self.dic
            for key in key_chunks[:-1]:
                if key not in entry:
                    entry[key] = {}
                entry = entry[key]
            entry[key_chunks[-1]] = value
        else:
            self.dic[key] = value

    def __contains__(self, item):
        if isinstance(item, str) and item[0] == "!":
            item_chunks = item[1:].split(".")
            entry = self.dic
            for item in item_chunks:
                if not isinstance(entry, dict) or item not in entry:
                    return False
                entry = entry[item]
            return True
        else:
            return item in self.dic

    def __repr__(self):
        msg = "<SystemDict> contents:"
        for key in self.dic.keys():
            val = self.dic[key]
            msg += "\n{}: ".format(key)
            if isinstance(val, dict):
                for subkey in val.keys():
                    msg += "\n  {}: {}".format(subkey, val[subkey])
            else:
                msg += "{}\n".format(val)
        return msg


def recursive_update(old_dict, new_dict):
    if new_dict is not None:
        for key in new_dict:
            if old_dict is not None and key in old_dict:
                if isinstance(old_dict[key], dict):
                    if isinstance(new_dict[key], dict):
                        old_dict[key] = recursive_update(old_dict[key],
                                                         new_dict[key])
                    else:
                        logging.warning("Overwriting dict: {} with non-dict: {}"
                                      "".format(old_dict[key], new_dict[key]))
                        old_dict[key] = new_dict[key]
                else:
                    if isinstance(new_dict[key], dict):
                        logging.warning("Overwriting non-dict: {} with dict: {}"
                                      "".format(old_dict[key], new_dict[key]))
                    old_dict[key] = new_dict[key]
            else:
                old_dict[key] = new_dict[key]

    return old_dict





from copy import deepcopy

import numpy as np
from astropy import units as u

from . import image_plane_utils as imp_utils
from .fov import FieldOfView
from .. import effects as efs
from ..effects.effects_utils import get_all_effects
from ..utils import check_keys


def get_3d_shifts(effects, **kwargs):
    """
    Returns the total 3D shifts (x,y,lam) from a series of Shift3D objects

    Parameters
    ----------
    effects : list of Shift3D effects

    Returns
    -------
    shift_dict : dict
        returns the x, y shifts for each wavelength in the fov_grid, where
        fov_grid contains the edge wavelengths for each spectral layer

    Notes
    -----
    Units returned by fov_grid():
    - wavelength: [um]
    - x_shift, y_shift: [deg]

    """
    required_keys = ["wave_min", "wave_mid", "wave_max",
                     "sub_pixel_fraction", "pixel_scale"]
    check_keys(kwargs, required_keys, action="warning")

    effects = get_all_effects(effects, efs.Shift3D)
    if len(effects) > 0:
        # shifts = [[waves], [x_shifts], [y_shifts]]
        shifts = [eff.fov_grid(which="shifts", **kwargs) for eff in effects]

        old_bin_edges = [shift[0] for shift in shifts if len(shift[0]) >= 2]
        new_bin_edges = np.unique(np.sort(np.concatenate(old_bin_edges),
                                          kind="stable"))

        x_shifts = np.zeros(len(new_bin_edges))
        y_shifts = np.zeros(len(new_bin_edges))
        # .. todo:: replace the 1e-7 with a variable in !SIM
        for shift in shifts:
            if not np.all(np.abs(shift[1]) < 1e-7):
                x_shifts += np.interp(new_bin_edges, shift[0], shift[1])
            if not np.all(np.abs(shift[2]) < 1e-7):
                y_shifts += np.interp(new_bin_edges, shift[0], shift[2])

        # After adding all the shifts together, work out a new wavelength set
        z_edges = np.copy(new_bin_edges)
        z_shifts = (x_shifts**2 + y_shifts**2)**0.5        # in arcsec
        step_size = kwargs["pixel_scale"] * kwargs["sub_pixel_fraction"]
        z_steps = (z_shifts / step_size).astype(int)
        # find where the shift is larger than a sub pixel fraction size
        ii = np.where(np.diff(z_steps) != 0)[0]

        x_shifts = np.array([x_shifts[0]] + list(x_shifts[ii]) + [x_shifts[-1]])
        y_shifts = np.array([y_shifts[0]] + list(y_shifts[ii]) + [y_shifts[-1]])
        z_edges  = np.array([z_edges[0]]  + list(z_edges[ii])  + [z_edges[-1]])

    else:
        z_edges = np.array([kwargs["wave_min"], kwargs["wave_max"]])
        x_shifts = np.zeros(2)
        y_shifts = np.zeros(2)

    shift_dict = {"wavelengths": z_edges,
                  "x_shifts": x_shifts / 3600.,   # fov_grid returns [arcsec]
                  "y_shifts": y_shifts / 3600.}   # get_3d_shifts returns [deg]

    return shift_dict


def get_imaging_waveset(effects_list, **kwargs):
    """
    Returns the edge wavelengths for the spectral layers needed for simulation

    Parameters
    ----------
    effects_list : list of Effect objects

    Returns
    -------
    wave_bin_edges : list
        [um] list of wavelengths

    """
    required_keys = ["wave_min", "wave_max"]
    check_keys(kwargs, required_keys, action="error")

    # get the filter wavelengths first to set (wave_min, wave_max)
    filters = get_all_effects(effects_list, (efs.FilterCurve, efs.FilterWheel))

    wave_bin_edges = [filt.fov_grid(which="waveset", **kwargs)
                      for filt in filters]
    if len(wave_bin_edges) > 0:
        kwargs["wave_min"] = np.max([w[0].value for w in wave_bin_edges])
        kwargs["wave_max"] = np.min([w[1].value for w in wave_bin_edges])
    wave_bin_edges = [[kwargs["wave_min"], kwargs["wave_max"]]]

    if kwargs["wave_min"] > kwargs["wave_max"]:
        raise ValueError("Filter wavelength ranges do not overlap: {}"
                         "".format(wave_bin_edges))

    # ..todo: add in Atmospheric dispersion and ADC here
    for effect_class in [efs.PSF]:
        effects = get_all_effects(effects_list, effect_class)
        for eff in effects:
            waveset = eff.fov_grid(which="waveset", **kwargs)
            if waveset is not None:
                wave_bin_edges += [waveset]

    wave_bin_edges = combine_wavesets(wave_bin_edges)

    if len(wave_bin_edges) == 0:
        wave_bin_edges = [kwargs["wave_min"], kwargs["wave_max"]]

    return wave_bin_edges


def get_imaging_headers(effects, **kwargs):
    """
    Returns a list of Header objects for each of the FieldOfVIew objects

    Parameters
    ----------
    effects : list of Effect objects
        Should contain all effects which return a header defining the extent
        of their spatial coverage

    Returns
    -------
    hdrs : list of Header objects

    Notes
    -----
    FOV headers use the return values from the ``<Effect>.fov_grid()``
    method. The ``fov_grid`` dict must contain the entry ``edges``

    This may change in future versions of ScopeSim

    """
    # look for apertures,
    #   if no apertures, look for detectors, make aperture from detector
    # get aperture headers via fov_grid()
    # check size of aperture in pixels
    #   if larger than max_chunk_size, split into smaller headers
    # add image plane WCS information for a direct projection

    required_keys = ["pixel_scale", "plate_scale",
                     "chunk_size", "max_segment_size"]
    check_keys(kwargs, required_keys, action="error")

    plate_scale = kwargs["plate_scale"]     # ["/mm]
    pixel_scale = kwargs["pixel_scale"]     # ["/pix]

    # look for apertures
    aperture_effects = get_all_effects(effects, (efs.ApertureMask,
                                                 efs.SlitWheel,
                                                 efs.ApertureList))
    if len(aperture_effects) == 0:
        detector_arrays = get_all_effects(effects, efs.DetectorList)
        if len(detector_arrays) > 0:
            aperture_effects += [detarr.fov_grid(which="edges",
                                                 pixel_scale=pixel_scale)
                                 for detarr in detector_arrays]
        else:
            raise ValueError("No ApertureMask or DetectorList was provided. At "
                             "least one must be passed to make an ImagePlane: "
                             "{}".format(effects))

    # get aperture headers from fov_grid()
    # - for-loop catches mutliple headers from ApertureList.fov_grid()
    hdrs = []
    for ap_eff in aperture_effects:
        # ..todo:: add this functionality to ApertureList effect
        hdr = ap_eff.fov_grid(which="edges", pixel_scale=pixel_scale)
        hdrs += hdr if isinstance(hdr, (list, tuple)) else [hdr]

    # check size of aperture in pixels - split if necessary
    sky_hdrs = []
    for hdr in hdrs:
        if hdr["NAXIS1"] * hdr["NAXIS2"] > kwargs["max_segment_size"]:
            sky_hdrs += imp_utils.split_header(hdr, kwargs["chunk_size"])
        else:
            sky_hdrs += [hdr]

    # ..todo:: Deal with the case that two or more ApertureMasks overlap

    # map the on-sky apertures directly to the image plane using plate_scale
    # - further changes can be made by the individual effects

    # plate_scale ["/mm] --> deg
    # pixel_scale ["/pix]
    # x_sky [deg] / (plate_scale/3600) [deg/mm] --> x_det [mm]

    pixel_size = pixel_scale / plate_scale
    plate_scale_deg = plate_scale / 3600.   # ["/mm] / 3600 = [deg/mm]
    for skyhdr in sky_hdrs:
        x_sky, y_sky = imp_utils.calc_footprint(skyhdr)
        x_det = x_sky / plate_scale_deg
        y_det = y_sky / plate_scale_deg

        dethdr = imp_utils.header_from_list_of_xy(x_det, y_det, pixel_size, "D")
        skyhdr.update(dethdr)

    return sky_hdrs


def get_imaging_fovs(headers, waveset, shifts, **kwargs):
    """
    Returns a list of ``FieldOfView`` objects

    Parameters
    ----------
    headers : list of fits.Header objects
        Headers giving spatial extent of each FOV region

    waveset : list of floats
        [um] N+1 wavelengths for N spectral layers

    shifts : list of tuples
        [deg] x,y shifts w.r.t to the optical axis plane. N shifts for N
        spectral layers

    Returns
    -------
    fovs : list of FieldOfView objects

    """

    shift_waves = shifts["wavelengths"]     # in [um]
    shift_dx = shifts["x_shifts"]           # in [deg]
    shift_dy = shifts["y_shifts"]

    # combine the wavelength bins from 1D spectral effects and 3D shift effects
    if len(shifts["wavelengths"]) > 0:
        mask = (shift_waves > np.min(waveset)) * (shift_waves < np.max(waveset))
        waveset = combine_wavesets([waveset, shift_waves[mask]])

    counter = 0
    fovs = []

    print("Preparing {} FieldOfViews".format((len(waveset)-1)*len(headers)),
          flush=True)

    for ii in range(len(waveset) - 1):
        for hdr in headers:
            # add any pre-instrument shifts to the FOV sky coords
            wave_mid = 0.5 * (waveset[ii] + waveset[ii+1])
            x_shift = np.interp(wave_mid, shift_waves, shift_dx)
            y_shift = np.interp(wave_mid, shift_waves, shift_dy)

            fov_hdr = deepcopy(hdr)
            fov_hdr["CRVAL1"] += x_shift      # headers are in [deg]
            fov_hdr["CRVAL2"] += y_shift

            # define the wavelength range for the FOV
            waverange = [waveset[ii], waveset[ii + 1]]

            # Make the FOV
            fov = FieldOfView(fov_hdr, waverange, id=counter, **kwargs)
            fovs += [fov]
            counter += 1

    return fovs


def get_spectroscopy_headers(effects, **kwargs):

    required_keys = ["pixel_scale", "plate_scale",
                     "wave_min", "wave_max"]
    check_keys(kwargs, required_keys, action="error")

    surface_list_effects = get_all_effects(effects, (efs.SurfaceList,
                                                     efs.FilterWheel,
                                                     efs.FilterCurve))
    detector_list_effects = get_all_effects(effects, efs.DetectorList)
    spec_trace_effects = get_all_effects(effects, efs.SpectralTraceList)
    aperture_effects = get_all_effects(effects, (efs.ApertureList,
                                                 efs.SlitWheel,
                                                 efs.ApertureMask))

    if len(surface_list_effects) > 0:
        waves = surface_list_effects[0].fov_grid(which="waveset")
        if len(waves) == 2:
            kwargs["wave_min"] = np.max([waves[0].value, kwargs["wave_min"]])
            kwargs["wave_max"] = np.min([waves[1].value, kwargs["wave_max"]])

    if len(detector_list_effects) > 0:
        implane_hdr = detector_list_effects[0].image_plane_header
    elif len(spec_trace_effects) > 0:
        implane_hdr = spec_trace_effects[0].image_plane_header
    else:
        raise ValueError("Missing a way to determine the image plane size")

    # ..todo: deal with multiple trace lists
    if len(spec_trace_effects) != 1:
        raise ValueError("More than one SpectralTraceList was found: {}"
                         "".format(spec_trace_effects))
    spec_trace = spec_trace_effects[0]

    sky_hdrs = []
    for ap_eff in aperture_effects:
        # if ApertureList, a list of ApertureMask headers is returned
        # If ApertureMask, a single header is returned
        hdr = ap_eff.fov_grid(which="edges", pixel_scale=kwargs["pixel_scale"])
        sky_hdrs += hdr if isinstance(hdr, (list, tuple)) else [hdr]

    fov_headers = [spec_trace.fov_grid(which="edges",
                                       sky_header=sky_hdr,
                                       det_header=implane_hdr,
                                       wave_min=kwargs["wave_min"],
                                       wave_max=kwargs["wave_max"],
                                       pixel_scale=kwargs["pixel_scale"],
                                       plate_scale=kwargs["plate_scale"]
                                       )
                   for sky_hdr in sky_hdrs]
    fov_headers = [hdr for hdr_list in fov_headers for hdr in hdr_list]
    # ..todo: check that each header is not larger than chunk_size

    return fov_headers


def get_spectroscopy_fovs(headers, shifts, effects=[], **kwargs):

    shift_waves = shifts["wavelengths"]     # in [um]
    shift_dx = shifts["x_shifts"]           # in [deg]
    shift_dy = shifts["y_shifts"]

    print("Preparing {} FieldOfViews".format(len(headers)), flush=True)

    apertures = get_all_effects(effects, (efs.ApertureList, efs.ApertureMask))
    masks = [ap.fov_grid(which="masks") for ap in apertures]
    mask_dict = {}
    for mask in masks:
        if isinstance(mask, dict):
            mask_dict.update(mask)
        elif isinstance(mask, np.ndarray):
            mask_dict[len(mask_dict)] = mask

    fovs = []
    for ii, hdr in enumerate(headers):
        # add any pre-instrument shifts to the FOV sky coords
        wave_mid = hdr["WAVE_MID"]
        x_shift = np.interp(wave_mid, shift_waves, shift_dx)
        y_shift = np.interp(wave_mid, shift_waves, shift_dy)

        fov_hdr = deepcopy(hdr)
        fov_hdr["CRVAL1"] += x_shift      # headers are in [deg]
        fov_hdr["CRVAL2"] += y_shift

        # Make the FOV
        fov = FieldOfView(fov_hdr, waverange=[hdr["WAVE_MIN"], hdr["WAVE_MAX"]],
                          **kwargs)
        fov.meta["distortion"]["rotation"] = hdr["ROTANGD"]
        fov.meta["distortion"]["shear"] = hdr["SKEWANGD"]
        fov.meta["conserve_image"] = hdr["IMG_CONS"]
        fov.meta["fov_id"] = ii
        fov.meta["aperture_id"] = hdr["APERTURE"]

        # .. todo: get these masks working
        # there needs to be fov_grid(which="mask") in ApertureList/Mask
        # fov.mask = mask_dict[hdr["APERTURE"]]
        fovs += [fov]

    return fovs


def combine_wavesets(waveset_list):
    """
    Joins and sorts several sets of wavelengths into a single 1D array

    Parameters
    ----------
    waveset_list : list
        A group of wavelength arrays or lists

    Returns
    -------
    wave_set : np.ndarray
        Combined set of wavelengths

    """
    wave_set = []
    for wbe in waveset_list:        # wbe = waveset bin edges
        wbe = wbe.value if isinstance(wbe, u.Quantity) else wbe
        wave_set += list(wbe)
    # ..todo:: set variable in !SIM.computing for rounding to the 7th decimal
    wave_set = np.unique(np.round(np.sort(wave_set, kind="stable"), 7))

    return wave_set
